from __future__ import annotations

import asyncio
import functools
import grp
import importlib
import importlib.resources
import ipaddress
import logging
import os
import pwd
import signal
import ssl
import sys
import traceback
import uuid
from collections.abc import AsyncGenerator
from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path
from typing import Any, AsyncIterator, Final, Iterable, Mapping, Sequence, cast

import aiohttp_cors
import aiohttp_jinja2
import aiomonitor
import aiotools
import click
import jinja2
import memray
import pyroscope
from aiohttp import web
from aiohttp.web_app import CleanupError
from setproctitle import setproctitle
from tenacity import AsyncRetrying, TryAgain, retry_if_exception_type, wait_exponential

from ai.backend.appproxy.common.config import get_default_redis_key_ttl
from ai.backend.appproxy.common.defs import (
    AGENTID_WORKER,
    APPPROXY_ANYCAST_STREAM_KEY,
    APPPROXY_BROADCAST_CHANNEL,
)
from ai.backend.appproxy.common.events import (
    AppProxyCircuitCreatedEvent,
    AppProxyCircuitRemovedEvent,
    AppProxyCircuitRouteUpdatedEvent,
    AppProxyWorkerCircuitAddedEvent,
)
from ai.backend.appproxy.common.exceptions import (
    BackendError,
    CoordinatorConnectionError,
    GenericBadRequest,
    GenericForbidden,
    InternalServerError,
    MethodNotAllowed,
    URLNotFound,
)
from ai.backend.appproxy.common.types import (
    AppCreator,
    EventLoopType,
    FrontendMode,
    FrontendServerMode,
    ProxyProtocol,
    WebMiddleware,
    WebRequestHandler,
)
from ai.backend.appproxy.common.utils import (
    BackendAIAccessLogger,
    ensure_json_serializable,
    mime_match,
)
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.defs import (
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    RedisRole,
)
from ai.backend.common.dto.internal.health import HealthResponse, HealthStatus
from ai.backend.common.events.dispatcher import EventDispatcher, EventHandler, EventProducer
from ai.backend.common.health_checker.checkers.valkey import ValkeyHealthChecker
from ai.backend.common.health_checker.probe import HealthProbe, HealthProbeOptions
from ai.backend.common.health_checker.types import ComponentId
from ai.backend.common.message_queue.hiredis_queue import HiRedisQueue
from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.message_queue.redis_queue import RedisMQArgs, RedisQueue
from ai.backend.common.metrics.http import build_api_metric_middleware
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.service_discovery.redis_discovery.service_discovery import (
    RedisServiceDiscovery,
    RedisServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscovery,
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)
from ai.backend.common.types import AgentId, RedisProfileTarget, ServiceDiscoveryType
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import OpenTelemetrySpec

from . import __version__
from .config import ServerConfig
from .config import load as load_config
from .coordinator_client import (
    deregister_worker,
    list_worker_circuits,
    ping_worker,
    register_worker,
)
from .metrics import collect_inference_metric
from .proxy.frontend import (
    H2PortFrontend,
    H2SubdomainFrontend,
    HTTPPortFrontend,
    HTTPSubdomainFrontend,
    TCPFrontend,
    TraefikPortFrontend,
    TraefikSubdomainFrontend,
    TraefikTCPFrontend,
)
from .types import Circuit, CleanupContext, RootContext, WorkerMetricRegistry

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

REDIS_APPPROXY_DB: Final[int] = 10  # FIXME: move to ai.backend.common.defs
EVENT_DISPATCHER_CONSUMER_GROUP: Final[str] = "appproxy-worker"

global_subapp_pkgs: Final[list[str]] = [
    ".health",
    ".setup",
]


@web.middleware
async def request_context_aware_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    request_id = request.headers.get("X-BackendAI-RequestID", str(uuid.uuid4()))
    request["request_id"] = request_id
    if _current_task := asyncio.current_task():
        setattr(_current_task, "request_id", request_id)
    resp = await handler(request)
    return resp


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    request_id = request.headers.get("X-BackendAI-RequestID", str(uuid.uuid4()))
    request["request_id"] = request_id
    if _current_task := asyncio.current_task():
        setattr(_current_task, "request_id", request_id)
    resp = await _handler(request)
    return resp


@web.middleware
async def exception_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]

    try:
        resp = await handler(request)
    except BackendError as ex:
        if ex.status_code == 500:
            log.exception("Internal server error raised inside handlers")
        if mime_match(request.headers.get("accept", "text/html"), "application/json", strict=True):
            return web.json_response(
                ensure_json_serializable(ex.body_dict),
                status=ex.status_code,
                headers={"Access-Control-Allow-Origin": "*"},
            )
        else:
            return aiohttp_jinja2.render_template(
                "error.jinja2",
                request,
                ex.body_dict,
                status=ex.status_code,
            )
    except web.HTTPException as ex:
        if ex.status_code == 404:
            raise URLNotFound(extra_data=request.path)
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(
                method=concrete_ex.method, allowed_methods=concrete_ex.allowed_methods
            )
        log.warning("Bad request: {0!r}", ex)
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
        raise e
    except Exception as e:
        log.exception("Uncaught exception in HTTP request handlers {0!r}", e)
        if root_ctx.local_config.debug.enabled:
            raise InternalServerError(traceback.format_exc())
        else:
            raise InternalServerError()
    else:
        return resp


async def request_counter_marker(root_ctx: RootContext) -> None:
    """Request counter marker function using the valkey client."""
    while True:
        try:
            redis_key = await root_ctx.request_counter_redis_queue.get()
            await root_ctx.valkey_live.incr_live_data(redis_key)
        except Exception:
            # log errors and keep going on
            log.exception("request_counter_marker(): error while handling request:")


async def last_used_time_marker(root_ctx: RootContext) -> None:
    """Last used time marker function using the valkey client."""
    while True:
        try:
            keys, last_used = await root_ctx.last_used_time_marker_redis_queue.get()
            data = {key: str(last_used) for key in keys}
            ttl = get_default_redis_key_ttl()
            await root_ctx.valkey_live.store_multiple_live_data(data, ex=ttl)
        except Exception:
            # log errors and keep going on
            log.exception("last_used_time_marker(): error while handling request:")


@asynccontextmanager
async def redis_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    redis_profile_target = RedisProfileTarget.from_dict(root_ctx.local_config.redis.to_dict())

    root_ctx.valkey_live = await ValkeyLiveClient.create(
        valkey_target=redis_profile_target.profile_target(RedisRole.LIVE).to_valkey_target(),
        db_id=REDIS_LIVE_DB,
        human_readable_name="appproxy-worker",
    )
    root_ctx.valkey_stat = await ValkeyStatClient.create(
        valkey_target=redis_profile_target.profile_target(RedisRole.STATISTICS).to_valkey_target(),
        db_id=REDIS_STATISTICS_DB,
        human_readable_name="appproxy-worker",
    )
    root_ctx.last_used_time_marker_redis_queue = asyncio.Queue()
    root_ctx.request_counter_redis_queue = asyncio.Queue()

    last_used_time_marker_task = asyncio.create_task(last_used_time_marker(root_ctx))
    request_counter_marker_task = asyncio.create_task(request_counter_marker(root_ctx))
    try:
        yield
    finally:
        last_used_time_marker_task.cancel()
        await last_used_time_marker_task
        request_counter_marker_task.cancel()
        await request_counter_marker_task

        await root_ctx.valkey_live.close()
        await root_ctx.valkey_stat.close()


@asynccontextmanager
async def http_client_pool_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from functools import partial

    import aiohttp

    from ai.backend.common.clients.http_client.client_pool import (
        ClientPool,
        tcp_client_session_factory,
    )

    client_timeout = aiohttp.ClientTimeout(
        total=None,
        connect=10.0,
        sock_connect=10.0,
        sock_read=None,
    )
    cleanup_interval = root_ctx.local_config.proxy_worker.client_pool_cleanup_interval

    root_ctx.http_client_pool = ClientPool(
        partial(
            tcp_client_session_factory,
            timeout=client_timeout,
            ssl=True,  # SSL verification per endpoint via ClientKey
        ),
        cleanup_interval_seconds=cleanup_interval,
    )
    try:
        yield
    finally:
        await root_ctx.http_client_pool.close()


async def _make_message_queue(
    root_ctx: RootContext,
) -> AbstractMessageQueue:
    node_id = root_ctx.local_config.proxy_worker.id
    redis_profile_target = RedisProfileTarget.from_dict(root_ctx.local_config.redis.to_dict())

    stream_redis_target = redis_profile_target.profile_target(RedisRole.STREAM)

    args = RedisMQArgs(
        anycast_stream_key=APPPROXY_ANYCAST_STREAM_KEY,
        broadcast_channel=APPPROXY_BROADCAST_CHANNEL,
        consume_stream_keys={
            APPPROXY_ANYCAST_STREAM_KEY,
        },
        subscribe_channels={
            APPPROXY_BROADCAST_CHANNEL,
        },
        group_name=EVENT_DISPATCHER_CONSUMER_GROUP,
        node_id=node_id,
        db=REDIS_STREAM_DB,
    )

    if root_ctx.local_config.proxy_worker.use_experimental_redis_event_dispatcher:
        return HiRedisQueue(
            stream_redis_target,
            args,
        )
    else:
        return await RedisQueue.create(
            stream_redis_target,
            args,
        )


@asynccontextmanager
async def event_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    mq = await _make_message_queue(root_ctx)
    root_ctx.event_producer = EventProducer(
        mq,
        source=AGENTID_WORKER,
        log_events=root_ctx.local_config.debug.log_events,
    )
    root_ctx.event_dispatcher = EventDispatcher(
        mq,
        log_events=root_ctx.local_config.debug.log_events,
        event_observer=root_ctx.metrics.event,
    )
    await root_ctx.event_dispatcher.start()
    try:
        yield
    finally:
        await root_ctx.event_producer.close()
        await asyncio.sleep(0.2)
        await root_ctx.event_dispatcher.close()


@asynccontextmanager
async def event_handler_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    handlers: list[EventHandler] = [
        root_ctx.event_dispatcher.subscribe(
            evt, root_ctx, handle_proxy_route_event, name="proxy-worker"
        )
        for evt in (
            AppProxyCircuitCreatedEvent,
            AppProxyCircuitRouteUpdatedEvent,
            AppProxyCircuitRemovedEvent,
        )
    ]
    try:
        yield
    finally:
        for handler in handlers:
            root_ctx.event_dispatcher.unsubscribe(handler)


@asynccontextmanager
async def proxy_frontend_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    match (
        root_ctx.local_config.proxy_worker.protocol,
        root_ctx.local_config.proxy_worker.frontend_mode,
    ):
        case (ProxyProtocol.HTTP, FrontendServerMode.TRAEFIK):
            assert root_ctx.local_config.proxy_worker.traefik
            match root_ctx.local_config.proxy_worker.traefik.frontend_mode:
                case FrontendMode.PORT:
                    root_ctx.proxy_frontend = TraefikPortFrontend(root_ctx)
                case FrontendMode.WILDCARD_DOMAIN:
                    root_ctx.proxy_frontend = TraefikSubdomainFrontend(root_ctx)
        case (ProxyProtocol.TCP, FrontendServerMode.TRAEFIK):
            assert (
                root_ctx.local_config.proxy_worker.traefik
                and root_ctx.local_config.proxy_worker.traefik.port_proxy
            )
            root_ctx.proxy_frontend = TraefikTCPFrontend(root_ctx)
        case (ProxyProtocol.HTTP, FrontendServerMode.PORT):
            root_ctx.proxy_frontend = HTTPPortFrontend(root_ctx)
        case (ProxyProtocol.HTTP, FrontendServerMode.WILDCARD_DOMAIN):
            root_ctx.proxy_frontend = HTTPSubdomainFrontend(root_ctx)
        case (ProxyProtocol.TCP, FrontendServerMode.PORT):
            root_ctx.proxy_frontend = TCPFrontend(root_ctx)
        case (ProxyProtocol.HTTP2, FrontendServerMode.PORT):
            root_ctx.proxy_frontend = H2PortFrontend(root_ctx)
        case (ProxyProtocol.HTTP2, FrontendServerMode.WILDCARD_DOMAIN):
            root_ctx.proxy_frontend = H2SubdomainFrontend(root_ctx)
        case _:
            log.error("Unsupported protocol {}", root_ctx.local_config.proxy_worker.protocol)
    await root_ctx.proxy_frontend.start()
    log.debug("started proxy protocol {}", root_ctx.proxy_frontend.__class__.__name__)
    try:
        yield
    finally:
        await root_ctx.proxy_frontend.terminate_all_circuits()
        try:
            await root_ctx.proxy_frontend.stop()
        except CleanupError as ee:
            if all([isinstance(e, asyncio.CancelledError) for e in ee.exceptions]):
                raise asyncio.CancelledError()
            else:
                raise ee


@asynccontextmanager
async def worker_registration_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    async for attempt in AsyncRetrying(
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(TryAgain),
    ):
        with attempt:
            try:
                await register_worker(root_ctx, str(uuid.uuid4()))
            except CoordinatorConnectionError:
                log.warning(
                    "Failed to connect to coordinator {}, retrying...",
                    root_ctx.local_config.proxy_worker.coordinator_endpoint,
                )
                raise TryAgain

    circuits = await list_worker_circuits(root_ctx, str(uuid.uuid4()))
    for circuit in circuits:
        await root_ctx.proxy_frontend.register_circuit(circuit, circuit.route_info)

    async def _heartbeat(interval: float):
        try:
            async for attempt in AsyncRetrying(
                wait=wait_exponential(multiplier=1, min=4, max=10),
                retry=retry_if_exception_type(TryAgain),
            ):
                with attempt:
                    try:
                        await ping_worker(root_ctx, str(uuid.uuid4()))
                    except CoordinatorConnectionError:
                        log.warning(
                            "Failed to ping coordinator {}, retrying...",
                            root_ctx.local_config.proxy_worker.coordinator_endpoint,
                        )

        except Exception as e:
            log.warning("Failed to ping coordinator: {}", str(e))

    timer = aiotools.create_timer(_heartbeat, root_ctx.local_config.proxy_worker.heartbeat_period)
    try:
        yield
    finally:
        await aiotools.cancel_and_wait(timer)
        await deregister_worker(root_ctx, str(uuid.uuid4()))


@asynccontextmanager
async def inference_metric_collection_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    timer = aiotools.create_timer(
        functools.partial(collect_inference_metric, root_ctx),
        root_ctx.local_config.proxy_worker.inference_metric_collection_interval,
    )
    try:
        yield
    finally:
        await aiotools.cancel_and_wait(timer)


@asynccontextmanager
async def health_probe_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """Initialize and start health probe with all health checkers."""
    probe = HealthProbe(options=HealthProbeOptions(check_interval=60))
    root_ctx.health_probe = probe

    # Register health checkers using already-initialized resources
    await probe.register(
        ValkeyHealthChecker(
            clients={
                ComponentId("live"): root_ctx.valkey_live,
                ComponentId("stat"): root_ctx.valkey_stat,
            }
        )
    )

    # Start periodic health checking
    await probe.start()
    try:
        yield
    finally:
        await probe.stop()


@asynccontextmanager
async def service_discovery_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    sd_type = root_ctx.local_config.service_discovery.type
    service_discovery: ServiceDiscovery
    match sd_type:
        case ServiceDiscoveryType.REDIS:
            redis_profile_target = RedisProfileTarget.from_dict(
                root_ctx.local_config.redis.to_dict()
            )
            live_redis_target = redis_profile_target.profile_target(RedisRole.LIVE)
            service_discovery = await RedisServiceDiscovery.create(
                RedisServiceDiscoveryArgs(valkey_target=live_redis_target.to_valkey_target())
            )
        case _:
            raise RuntimeError(f"Unsupported service discovery type: {sd_type}")

    # Determine announce addresses
    announce_addr = root_ctx.local_config.proxy_worker.announce_addr
    assert announce_addr is not None  # auto-populated if None
    sd_loop = ServiceDiscoveryLoop(
        sd_type,
        service_discovery,
        ServiceMetadata(
            display_name=f"appproxy-worker-{root_ctx.local_config.proxy_worker.authority}",
            service_group="appproxy-worker",
            version=__version__,
            endpoint=ServiceEndpoint(
                address=announce_addr.host,
                port=announce_addr.port,
                protocol="http",
                # It can be separated into an internal-purpose port later.
                prometheus_address=str(announce_addr),
            ),
        ),
    )

    if root_ctx.local_config.otel.enabled:
        meta = sd_loop.metadata
        otel_spec = OpenTelemetrySpec(
            service_id=meta.id,
            service_name=meta.service_group,
            service_version=meta.version,
            log_level=root_ctx.local_config.otel.log_level,
            endpoint=root_ctx.local_config.otel.endpoint,
        )
        BraceStyleAdapter.apply_otel(otel_spec)
    try:
        yield
    finally:
        sd_loop.close()


async def metrics(request: web.Request) -> web.Response:
    request["do_not_print_access_log"] = True
    root_ctx: RootContext = request.app["_root.context"]
    allowed_network = ipaddress.IPv4Network(
        root_ctx.local_config.proxy_worker.metric_access_allowed_hosts
    )
    if not request.remote:
        raise GenericForbidden
    try:
        remote_ip = ipaddress.IPv4Network(request.remote)
        if not remote_ip.subnet_of(allowed_network):
            raise GenericForbidden
    except ValueError:
        raise GenericForbidden

    response = web.Response(
        text=root_ctx.metrics.to_prometheus(),
        content_type="text/plain",
    )
    return response


async def hello(request: web.Request) -> web.Response:
    """Health check endpoint with dependency connectivity status"""
    request["do_not_print_access_log"] = True
    root_ctx: RootContext = request.app["_root.context"]
    connectivity = await root_ctx.health_probe.get_connectivity_status()
    response = HealthResponse(
        status=HealthStatus.OK if connectivity.overall_healthy else HealthStatus.DEGRADED,
        version=__version__,
        component="appproxy-worker",
        connectivity=connectivity,
    )
    return web.json_response(response.model_dump(mode="json"))


async def status(request: web.Request) -> web.Response:
    request["do_not_print_access_log"] = True
    return web.json_response({"api_version": "v2"})


async def handle_proxy_route_event(
    context: RootContext,
    agent_id: AgentId,
    event: AppProxyCircuitCreatedEvent
    | AppProxyCircuitRouteUpdatedEvent
    | AppProxyCircuitRemovedEvent,
) -> None:
    if event.target_worker_authority != context.local_config.proxy_worker.authority:
        return

    match event:
        case AppProxyCircuitCreatedEvent():
            log.debug(
                "handle_proxy_route_event(evt: AppProxyCircuitCreatedEvent({}))",
                [c.id for c in event.circuits],
            )
            for circuit in event.circuits:
                await context.proxy_frontend.register_circuit(
                    Circuit.from_serialized_circuit(circuit), circuit.route_info
                )
            await context.event_producer.broadcast_event(
                AppProxyWorkerCircuitAddedEvent(
                    target_worker_authority=context.local_config.proxy_worker.authority,
                    circuits=event.circuits,
                )
            )
        case AppProxyCircuitRouteUpdatedEvent():
            log.debug(
                "handle_proxy_route_event(evt: AppProxyCircuitRouteUpdatedEvent({}))",
                event.circuit.id,
            )
            await context.proxy_frontend.update_circuit_route_info(
                Circuit.from_serialized_circuit(event.circuit), event.routes
            )
        case AppProxyCircuitRemovedEvent():
            log.debug(
                "handle_proxy_route_event(evt: AppProxyCircuitRemovedEvent({}))",
                [c.id for c in event.circuits],
            )
            for circuit in event.circuits:
                await context.proxy_frontend.break_circuit(Circuit.from_serialized_circuit(circuit))


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


def handle_loop_error(
    root_ctx: RootContext,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[WebMiddleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["_root.context"] = root_app["_root.context"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)  # type: ignore


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    root_ctx: RootContext = root_app["_root.context"]
    subapp, global_middlewares = create_subapp(root_ctx.cors_options)
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def build_root_app(
    pidx: int,
    local_config: ServerConfig,
    *,
    cleanup_contexts: Sequence[CleanupContext] | None = None,
    subapp_pkgs: Sequence[str] = [],
) -> web.Application:
    root_ctx = RootContext()
    root_ctx.metrics = WorkerMetricRegistry.instance()

    app = web.Application(
        middlewares=[
            request_context_aware_middleware,
            exception_middleware,
            api_middleware,
            build_api_metric_middleware(root_ctx.metrics.api),
        ]
    )
    global_exception_handler = functools.partial(handle_loop_error, root_ctx)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app["_root.context"] = root_ctx
    root_ctx.local_config = local_config
    root_ctx.pidx = pidx
    root_ctx.cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    app.on_response_prepare.append(on_prepare)

    with importlib.resources.as_file(importlib.resources.files("ai.backend.appproxy.common")) as f:
        template_path = f / "templates"
        aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

    if cleanup_contexts is None:
        if local_config.proxy_worker.frontend_mode == FrontendServerMode.TRAEFIK:
            # traefik won't require event dispatcher
            cleanup_contexts = [
                proxy_frontend_ctx,
                redis_ctx,
                http_client_pool_ctx,
                health_probe_ctx,
                service_discovery_ctx,
                worker_registration_ctx,
                inference_metric_collection_ctx,
            ]
        else:
            cleanup_contexts = [
                proxy_frontend_ctx,
                redis_ctx,
                http_client_pool_ctx,
                event_dispatcher_ctx,
                event_handler_ctx,
                health_probe_ctx,
                service_discovery_ctx,
                worker_registration_ctx,
                inference_metric_collection_ctx,
            ]
    shutdown_context_instances = []

    async def _cleanup_context_wrapper(app: web.Application) -> AsyncIterator[None]:
        # aiohttp's cleanup contexts are just async generators, not async context managers.
        assert cleanup_contexts is not None
        async with AsyncExitStack() as stack:
            for cctx in cleanup_contexts:
                cctx_instance = cctx(root_ctx)
                if hasattr(cctx_instance, "shutdown"):
                    shutdown_context_instances.append(cctx_instance)
                await stack.enter_async_context(cctx_instance)
            yield

    async def _trigger_shutdown(app: web.Application) -> None:
        # shutdown is triggered before cleanup, giving chances to close client connections first.
        for cctx_instance in shutdown_context_instances:
            try:
                await cctx_instance.shutdown()
            except Exception:
                log.exception("error while shutting down a cleanup context")

    app.on_shutdown.append(_trigger_shutdown)
    app.cleanup_ctx.append(_cleanup_context_wrapper)
    cors = aiohttp_cors.setup(app, defaults=root_ctx.cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))
    cors.add(app.router.add_route("GET", "/status", status))
    cors.add(app.router.add_route("GET", "/metrics", metrics))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.appproxy.worker.api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))
    return app


@asynccontextmanager
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncIterator[None]:
    local_config: ServerConfig = _args[0]
    loop.set_debug(local_config.debug.asyncio)
    root_app = build_root_app(pidx, local_config, subapp_pkgs=global_subapp_pkgs)

    @asynccontextmanager
    async def aiomonitor_ctx() -> AsyncGenerator[None]:
        # Start aiomonitor.
        m = aiomonitor.Monitor(
            loop,
            host="0.0.0.0",
            termui_port=local_config.proxy_worker.aiomonitor_termui_port + pidx,
            webui_port=local_config.proxy_worker.aiomonitor_webui_port + pidx,
            console_enabled=False,
            hook_task_factory=local_config.debug.enhanced_aiomonitor_task_info,
        )
        m.prompt = f"monitor (proxy-worker[{pidx}@{os.getpid()}]) >>> "
        # Add some useful console_locals for ease of debugging
        m.console_locals["root_app"] = root_app
        m.console_locals["root_ctx"] = root_app["_root.context"]
        aiomon_started = False
        try:
            m.start()
            aiomon_started = True
        except Exception as e:
            log.warning(
                "aiomonitor could not start but skipping this error to continue", exc_info=e
            )
        try:
            yield
        finally:
            if aiomon_started:
                m.close()

    @asynccontextmanager
    async def webapp_ctx() -> AsyncGenerator[None]:
        ssl_ctx = None
        if local_config.proxy_worker.tls_listen:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(local_config.proxy_worker.tls_cert),
                str(local_config.proxy_worker.tls_privkey),
            )

        runner = web.AppRunner(
            root_app, keepalive_timeout=30.0, access_log_class=BackendAIAccessLogger
        )
        await runner.setup()
        service_addr = local_config.proxy_worker.api_bind_addr
        site = web.TCPSite(
            runner,
            str(service_addr.host),
            service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        await site.start()
        try:
            yield
        finally:
            await runner.cleanup()

    worker_init_stack = AsyncExitStack()
    await worker_init_stack.__aenter__()
    try:
        await worker_init_stack.enter_async_context(aiomonitor_ctx())
        await worker_init_stack.enter_async_context(webapp_ctx())

        if os.geteuid() == 0:
            uid = local_config.proxy_worker.user
            gid = local_config.proxy_worker.group
            os.setgroups([
                g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem
            ])
            os.setgid(gid)
            os.setuid(uid)
            log.info("changed process uid and gid to {}:{}", uid, gid)

        log.info("Started the app-proxy worker service.")
    except Exception:
        log.exception("Server initialization failure; triggering shutdown...")
        loop.call_later(0.2, os.kill, 0, signal.SIGINT)
    try:
        yield
    finally:
        log.info("shutting down...")
        await worker_init_stack.__aexit__(None, None, None)


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: proxy-worker worker-{pidx}")
    local_config: ServerConfig = _args[0]
    log_endpoint: str = _args[1]
    logger = Logger(
        local_config.logging,
        is_master=False,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, _args):
                yield
    except Exception:
        traceback.print_exc(file=sys.stderr)


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help=(
        "The config file path. (default: ./proxy-worker.toml and /etc/backend.ai/proxy-worker.toml)"
    ),
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(ctx: click.Context, config_path: Path, debug: bool, log_level: LogLevel) -> None:
    """
    Start the proxy-worker service as a foreground process.
    """
    log_level = LogLevel.DEBUG if debug else log_level
    server_config = load_config(config_path, log_level)

    if ctx.invoked_subcommand is None:
        tracker: memray.Tracker | None = None
        if server_config.profiling.enable_pyroscope:
            assert server_config.profiling.pyroscope_config
            pyroscope.configure(**server_config.profiling.pyroscope_config.model_dump())
        if server_config.profiling.enable_memray:
            tracker = memray.Tracker(
                server_config.profiling.memray_output_destination,
                follow_fork=True,
            )
            tracker.__enter__()
        server_config.proxy_worker.pid_file.touch(exist_ok=True)
        server_config.proxy_worker.pid_file.write_text(str(os.getpid()))
        ipc_base_path = server_config.proxy_worker.ipc_base_path
        ipc_base_path.mkdir(exist_ok=True, parents=True)
        log_sockpath = ipc_base_path / f"worker-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                server_config.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                setproctitle("backend.ai: proxy-worker")
                log.info("Backend.AI AppProxy Worker {0}", __version__)
                log.info("runtime: {0}", env_info())
                if server_config.profiling.enable_pyroscope:
                    log.info("Pyroscope tracing enabled")
                if server_config.profiling.enable_memray:
                    log.info("Memray tracing enabled")
                log_config = logging.getLogger("ai.backend.appproxy.worker.config")
                log_config.debug("debug mode enabled.")
                match server_config.proxy_worker.event_loop:
                    case EventLoopType.UVLOOP:
                        import uvloop

                        runner = uvloop.run
                        log.info("Using uvloop as the event loop backend")
                    case EventLoopType.ASYNCIO:
                        runner = asyncio.run
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=1,
                        args=(server_config, log_endpoint),
                        wait_timeout=5.0,
                        runner=runner,
                    )
                finally:
                    log.info("terminated.")
        finally:
            if server_config.proxy_worker.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                server_config.proxy_worker.pid_file.unlink()
            if tracker:
                tracker.__exit__(None, None, None)
    else:
        # Click is going to invoke a subcommand.
        pass


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import asyncio
import dataclasses
import enum
import time
from dataclasses import dataclass
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Final,
    FrozenSet,
    Generic,
    Mapping,
    Optional,
    Self,
    TypeAlias,
    TypeVar,
    Union,
)
from uuid import UUID

import aiohttp_cors
import attrs
import prometheus_client

from ai.backend.appproxy.common.types import (
    AppMode,
    FrontendMode,
    SerializableCircuit,
)
from ai.backend.common.clients.http_client.client_pool import ClientPool
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.health_checker.probe import HealthProbe
from ai.backend.common.metrics.metric import (
    APIMetricObserver,
    EventMetricObserver,
    SystemMetricObserver,
)
from ai.backend.common.types import (
    MetricKey,
    MetricValue,
    MovingStatValue,
    RuntimeVariant,
)

if TYPE_CHECKING:
    from .config import ServerConfig
    from .proxy.frontend.base import BaseFrontend


class ProxyMetricObserver:
    _instance: Optional[Self] = None

    _upstream_request_sent_http: prometheus_client.Counter
    _upstream_response_received_http: prometheus_client.Counter
    _requests_received_http: prometheus_client.Counter
    _responses_sent_http: prometheus_client.Counter
    _pending_requests_http: prometheus_client.Gauge
    _proxy_request_iteration_duration_http: prometheus_client.Histogram
    _request_body_size_http: prometheus_client.Histogram
    _response_body_size_http: prometheus_client.Histogram

    _connections_established_ws: prometheus_client.Counter
    _connections_closed_ws: prometheus_client.Counter
    _pending_connections_ws: prometheus_client.Gauge
    _proxy_request_iteration_duration_ws: prometheus_client.Histogram
    _connection_processed_traffics_ws: prometheus_client.Counter
    _connection_total_traffics_ws: prometheus_client.Histogram

    _connections_established_tcp: prometheus_client.Counter
    _connections_closed_tcp: prometheus_client.Counter
    _pending_connections_tcp: prometheus_client.Gauge
    _proxy_request_iteration_duration_tcp: prometheus_client.Histogram
    _connection_processed_traffics_tcp: prometheus_client.Counter
    _connection_total_traffics_tcp: prometheus_client.Histogram

    def __init__(self) -> None:
        self._upstream_request_sent_http = prometheus_client.Counter(
            name="appproxy_upstream_request_sent_http",
            labelnames=["remote"],
            documentation="Total number of HTTP requests sent to upstream",
        )
        self._upstream_request_sent_http.labels(remote="").inc(0)
        self._upstream_response_received_http = prometheus_client.Counter(
            name="appproxy_upstream_response_received_http",
            labelnames=["remote"],
            documentation="Total number of HTTP responses received from upstream",
        )
        self._upstream_response_received_http.labels(remote="").inc(0)
        self._requests_received_http = prometheus_client.Counter(
            name="appproxy_requests_received_http",
            labelnames=["remote"],
            documentation="Total number of HTTP requests received from downstream",
        )
        self._requests_received_http.labels(remote="").inc(0)
        self._responses_sent_http = prometheus_client.Counter(
            name="appproxy_responses_sent_http",
            labelnames=["remote"],
            documentation="Total number of HTTP responses sent to downstream",
        )
        self._responses_sent_http.labels(remote="").inc(0)
        self._pending_requests_http = prometheus_client.Gauge(
            name="appproxy_pending_requests_http",
            labelnames=["remote"],
            documentation="Ongoing HTTP proxy requests initiated by downstream",
        )
        self._pending_requests_http.labels(remote="").set(0)
        self._proxy_request_iteration_duration_http = prometheus_client.Histogram(
            name="appproxy_proxy_request_iteration_duration_http",
            labelnames=["remote"],
            documentation="Total seconds taken to complete each HTTP proxy request",
        )
        self._request_body_size_http = prometheus_client.Histogram(
            name="appproxy_request_body_size_http",
            labelnames=["remote"],
            documentation="Request body size measured for each HTTP proxy request",
        )
        self._response_body_size_http = prometheus_client.Histogram(
            name="appproxy_response_body_size_http",
            labelnames=["remote"],
            documentation="Byte length of the response body generated from upstream HTTP response",
        )
        self._connections_established_ws = prometheus_client.Counter(
            name="appproxy_connections_established_ws",
            documentation="Total number of WebSocket connections established",
        )
        self._connections_established_ws.inc(0)
        self._connections_closed_ws = prometheus_client.Counter(
            name="appproxy_connections_closed_ws",
            documentation="Total number of WebSocket connections closed",
        )
        self._connections_closed_ws.inc(0)
        self._pending_connections_ws = prometheus_client.Gauge(
            name="appproxy_pending_connections_ws",
            documentation="Ongoing WebSocket proxy connections",
        )
        self._pending_connections_ws.set(0)
        self._proxy_request_iteration_duration_ws = prometheus_client.Histogram(
            name="appproxy_proxy_request_iteration_duration_ws",
            documentation="Total seconds taken to complete each WebSocket proxy request",
        )
        self._connection_processed_traffics_ws = prometheus_client.Counter(
            name="appproxy_connection_processed_traffics_ws",
            documentation="Number of bytes transferred from each WebSocket connection bidirectionally, updated on-the-fly",
        )
        self._connection_processed_traffics_ws.inc(0)
        self._connection_total_traffics_ws = prometheus_client.Histogram(
            name="appproxy_connection_total_traffics_ws",
            documentation="Number of bytes transferred from each WebSocket connection bidirectionally, updated after WebSocket connection closes",
        )
        self._connections_established_tcp = prometheus_client.Counter(
            name="appproxy_connections_established_tcp",
            documentation="Total number of TCP connections established",
        )
        self._connections_established_tcp.inc(0)
        self._connections_closed_tcp = prometheus_client.Counter(
            name="appproxy_connections_closed_tcp",
            documentation="Total number of TCP connections closed",
        )
        self._connections_closed_tcp.inc(0)
        self._pending_connections_tcp = prometheus_client.Gauge(
            name="appproxy_pending_connections_tcp",
            documentation="Ongoing TCP proxy connections",
        )
        self._pending_connections_tcp.set(0)
        self._proxy_request_iteration_duration_tcp = prometheus_client.Histogram(
            name="appproxy_proxy_request_iteration_duration_tcp",
            documentation="Total seconds taken to complete each TCP proxy request",
        )
        self._connection_processed_traffics_tcp = prometheus_client.Counter(
            name="appproxy_connection_processed_traffics_tcp",
            documentation="Number of bytes transferred from each TCP connection bidirectionally, updated on-the-fly",
        )
        self._connection_processed_traffics_tcp.inc(0)
        self._connection_total_traffics_tcp = prometheus_client.Histogram(
            name="appproxy_connection_total_traffics_tcp",
            documentation="Number of bytes transferred from each TCP connection bidirectionally, updated on-the-fly",
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_upstream_http_request(self, *, remote: str, total_bytes_size: int) -> None:
        self._upstream_request_sent_http.labels(remote=remote).inc()
        self._request_body_size_http.labels(remote=remote).observe(total_bytes_size)

    def observe_upstream_http_response(self, *, remote: str, total_bytes_size: int) -> None:
        self._upstream_response_received_http.labels(remote=remote).inc()
        self._response_body_size_http.labels(remote=remote).observe(total_bytes_size)

    def observe_downstream_http_request(self, *, remote: str) -> None:
        self._requests_received_http.labels(remote=remote).inc()
        self._pending_requests_http.labels(remote=remote).inc()

    def observe_downstream_http_response(self, *, remote: str, duration: int) -> None:
        self._pending_requests_http.labels(remote=remote).dec()
        self._proxy_request_iteration_duration_http.labels(remote=remote).observe(duration)
        self._responses_sent_http.labels(remote=remote).inc()

    def observe_upstream_ws_traffic_chunk(self, *, total_bytes_size: int) -> None:
        self._connection_processed_traffics_ws.inc(total_bytes_size)

    def observe_upstream_ws_connection_start(self) -> None:
        self._connections_established_ws.inc()
        self._pending_connections_ws.inc()

    def observe_upstream_ws_connection_end(self, *, duration: int, total_bytes_size: int) -> None:
        self._connections_closed_ws.inc()
        self._pending_connections_ws.dec()
        self._proxy_request_iteration_duration_ws.observe(duration)
        self._connection_total_traffics_ws.observe(total_bytes_size)

    def observe_upstream_tcp_traffic_chunk(self, total_bytes_size: int) -> None:
        self._connection_processed_traffics_tcp.inc(total_bytes_size)

    def observe_downstream_tcp_start(self) -> None:
        self._connections_established_tcp.inc()
        self._pending_connections_tcp.inc()

    def observe_downstream_tcp_end(self, *, duration: int) -> None:
        self._connections_closed_tcp.inc()
        self._pending_connections_tcp.dec()
        self._proxy_request_iteration_duration_tcp.observe(duration)


class CircuitMetricObserver:
    _instance: Optional[Self] = None

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    _circuits_created: prometheus_client.Counter
    _alive_circuits: prometheus_client.Gauge

    def __init__(self) -> None:
        self._circuits_created = prometheus_client.Counter(
            name="appproxy_circuits_created",
            labelnames=["protocol"],
            documentation="Total number of AppProxy circuits created on this Worker. `endpoint_id` is provided only when circuit is INFERENCE type. `user_id` is provided only when circuit is INTERACTIVE type.",
        )
        self._circuits_created.labels(protocol="").inc(0)
        self._alive_circuits = prometheus_client.Gauge(
            name="appproxy_alive_circuits",
            labelnames=["protocol"],
            documentation="Total number of AppProxy circuits created on this Worker. `endpoint_id` is provided only when circuit is INFERENCE type. `user_id` is provided only when circuit is INTERACTIVE type.",
        )
        self._alive_circuits.labels(protocol="").set(0)

    def observe_circuit_creation(self, *, protocol: str) -> None:
        self._circuits_created.labels(protocol=protocol).inc()
        self._alive_circuits.labels(protocol=protocol).inc()

    def observe_circuit_removal(self, *, protocol: str) -> None:
        self._alive_circuits.labels(protocol=protocol).dec()


class WorkerMetricRegistry:
    _instance: Optional[Self] = None

    api: APIMetricObserver
    proxy: ProxyMetricObserver
    circuit: CircuitMetricObserver
    event: EventMetricObserver
    system: SystemMetricObserver

    def __init__(self) -> None:
        self.api = APIMetricObserver.instance()
        self.proxy = ProxyMetricObserver.instance()
        self.circuit = CircuitMetricObserver.instance()
        self.event = EventMetricObserver.instance()
        self.system = SystemMetricObserver.instance()

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def to_prometheus(self) -> str:
        self.system.observe()
        return prometheus_client.generate_latest().decode("utf-8")


@dataclass
class PrometheusMetrics:
    registry: prometheus_client.CollectorRegistry

    redis_events_received: prometheus_client.Counter
    alive_asyncio_tasks: prometheus_client.Gauge
    memory_used_rss: prometheus_client.Gauge
    memory_used_vms: prometheus_client.Gauge


@attrs.define(slots=True, auto_attribs=True, init=False)
class RootContext:
    pidx: int
    proxy_frontend: BaseFrontend
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    valkey_live: ValkeyLiveClient
    valkey_stat: ValkeyStatClient
    http_client_pool: ClientPool
    worker_id: UUID
    local_config: ServerConfig
    last_used_time_marker_redis_queue: asyncio.Queue[tuple[list[str], float]]
    request_counter_redis_queue: asyncio.Queue[str]
    cors_options: dict[str, aiohttp_cors.ResourceOptions]
    metrics: WorkerMetricRegistry
    health_probe: HealthProbe


CleanupContext: TypeAlias = Callable[["RootContext"], AsyncContextManager[None]]
TCircuitKey = TypeVar("TCircuitKey", int, str)


@dataclass
class PortFrontendInfo:
    port: int


@dataclass
class SubdomainFrontendInfo:
    subdomain: str


@dataclass
class InteractiveAppInfo:
    user_id: UUID


@dataclass
class InferenceAppInfo:
    endpoint_id: UUID
    runtime_variant: RuntimeVariant | None


class Circuit(SerializableCircuit):
    frontend: PortFrontendInfo | SubdomainFrontendInfo

    port: int | None
    "for initialization usage only; use `frontend` variable"
    subdomain: str | None
    "for initialization usage only; use `frontend` variable"

    app_info: InteractiveAppInfo | InferenceAppInfo

    user_id: UUID | None
    "for initialization usage only; use `app_info` variable"
    endpoint_id: UUID | None
    "for initialization usage only; use `app_info` variable"
    runtime_variant: str | None
    "for initialization usage only; use `app_info` variable"

    _app_inference_metrics: dict[MetricKey, "Metric | HistogramMetric"]
    _replica_inference_metrics: dict[
        MetricKey, dict[UUID, "Metric | HistogramMetric"]
    ]  # [Metric Key:[Route id: Metric]] pair

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._app_inference_metrics = {}
        self._replica_inference_metrics = {}

    @classmethod
    def from_serialized_circuit(cls, circuit: SerializableCircuit) -> "Circuit":
        frontend: PortFrontendInfo | SubdomainFrontendInfo
        app_info: InteractiveAppInfo | InferenceAppInfo

        match circuit.app_mode:
            case AppMode.INTERACTIVE:
                assert circuit.user_id
                app_info = InteractiveAppInfo(circuit.user_id)
            case AppMode.INFERENCE:
                assert circuit.endpoint_id
                app_info = InferenceAppInfo(
                    circuit.endpoint_id,
                    RuntimeVariant(circuit.runtime_variant)
                    if circuit.runtime_variant is not None
                    else None,
                )

        match circuit.frontend_mode:
            case FrontendMode.PORT:
                assert circuit.port is not None
                frontend = PortFrontendInfo(circuit.port)
            case FrontendMode.WILDCARD_DOMAIN:
                assert circuit.subdomain is not None
                frontend = SubdomainFrontendInfo(circuit.subdomain)

        return cls(frontend=frontend, app_info=app_info, **circuit.model_dump())

    @property
    def prometheus_metric_label(self) -> dict[str, str]:
        metric_labels = {"protocol": self.protocol.name}

        return metric_labels


class MetricTypes(enum.Enum):
    """
    Specifies the type of a metric value.

    Currently this DOES NOT affect calculation and processing of the metric,
    but serves as a metadata for code readers.
    The actual calculation and formatting is controlled by :meth:`Metric.current_hook()`,
    :attr:`Metric.unit_hint` and :attr:`Metric.stats_filter`.
    """

    GAUGE = 0
    """
    Represents an instantly measured occupancy value.
    (e.g., used space as bytes, occupied amount as the number of items or a bandwidth)
    """

    # Enum value 1 and 2 intentionally left unused to match consistency with agent statistics

    ACCUMULATION = 3
    """
    Represents an accumulated value
    (e.g., total number of events, total period of occupation)
    """
    HISTOGRAM = 4
    """
    Represents a histogram of values
    (e.g., latency distribution)
    """


@dataclass
class Measurement:
    value: Decimal
    capacity: Optional[Decimal] = dataclasses.field(default=None)


@dataclass
class HistogramMeasurement:
    buckets: Mapping[str, Decimal]
    count: Optional[int] = dataclasses.field(default=None)
    sum: Optional[Decimal] = dataclasses.field(default=None)


TMeasurement = TypeVar("TMeasurement", bound=Union[Measurement, HistogramMeasurement])


@dataclass
class InferenceMeasurement(Generic[TMeasurement]):
    """
    Collection of per-inference framework statistics for a specific metric.
    """

    key: MetricKey
    type: MetricTypes
    per_app: TMeasurement
    per_replica: Mapping[UUID, TMeasurement]  # [Route Id: Measurement] pair
    stats_filter: FrozenSet[str] = dataclasses.field(default_factory=frozenset)
    unit_hint: str = dataclasses.field(default="count")


def remove_exponent(num: Decimal) -> Decimal:
    return num.quantize(Decimal(1)) if num == num.to_integral() else num.normalize()


class MovingStatistics:
    __slots__ = (
        "_sum",
        "_count",
        "_min",
        "_max",
        "_last",
    )
    _sum: Decimal
    _count: int
    _min: Decimal
    _max: Decimal
    _last: list[tuple[Decimal, float]]

    def __init__(self, initial_value: Optional[Decimal] = None):
        self._last = []
        if initial_value is None:
            self._sum = Decimal(0)
            self._min = Decimal("inf")
            self._max = Decimal("-inf")
            self._count = 0
        else:
            self._sum = initial_value
            self._min = initial_value
            self._max = initial_value
            self._count = 1
            point = (initial_value, time.perf_counter())
            self._last.append(point)

    def update(self, value: Decimal):
        self._sum += value
        self._min = min(self._min, value)
        self._max = max(self._max, value)
        self._count += 1
        point = (value, time.perf_counter())
        self._last.append(point)
        # keep only the latest two data points
        if len(self._last) > 2:
            self._last.pop(0)

    @property
    def min(self) -> Decimal:
        return self._min

    @property
    def max(self) -> Decimal:
        return self._max

    @property
    def sum(self) -> Decimal:
        return self._sum

    @property
    def avg(self) -> Decimal:
        return self._sum / self._count

    @property
    def diff(self) -> Decimal:
        if len(self._last) == 2:
            return self._last[-1][0] - self._last[-2][0]
        return Decimal(0)

    @property
    def rate(self) -> Decimal:
        if len(self._last) == 2:
            return (self._last[-1][0] - self._last[-2][0]) / Decimal(
                self._last[-1][1] - self._last[-2][1]
            )
        return Decimal(0)

    def to_serializable_dict(self) -> MovingStatValue:
        q = Decimal("0.000")
        return {
            "min": str(remove_exponent(self.min.quantize(q))),
            "max": str(remove_exponent(self.max.quantize(q))),
            "sum": str(remove_exponent(self.sum.quantize(q))),
            "avg": str(remove_exponent(self.avg.quantize(q))),
            "diff": str(remove_exponent(self.diff.quantize(q))),
            "rate": str(remove_exponent(self.rate.quantize(q))),
            "version": 2,
        }


@attrs.define(auto_attribs=True, slots=True)
class Metric:
    key: str
    type: MetricTypes
    unit_hint: str
    stats: MovingStatistics
    stats_filter: FrozenSet[str]
    current: Decimal
    capacity: Optional[Decimal] = None
    current_hook: Optional[Callable[["Metric"], Decimal]] = None

    def update(self, value: Measurement):
        if value.capacity is not None:
            self.capacity = value.capacity
        self.stats.update(value.value)
        self.current = value.value
        if self.current_hook is not None:
            self.current = self.current_hook(self)

    def to_serializable_dict(self) -> MetricValue:
        q = Decimal("0.000")
        q_pct = Decimal("0.00")
        return {
            "__type": self.type.name,  # type: ignore
            "current": str(remove_exponent(self.current.quantize(q))),
            "capacity": (
                str(remove_exponent(self.capacity.quantize(q)))
                if self.capacity is not None
                else None
            ),
            "pct": (
                str(
                    remove_exponent(
                        (Decimal(self.current) / Decimal(self.capacity) * 100).quantize(q_pct)
                    )
                )
                if (self.capacity is not None and self.capacity.is_normal() and self.capacity > 0)
                else "0.00"
            ),
            "unit_hint": self.unit_hint,
            **{
                f"stats.{k}": v  # type: ignore
                for k, v in self.stats.to_serializable_dict().items()
                if k in self.stats_filter
            },
        }


@attrs.define(auto_attribs=True, slots=True)
class HistogramMetric:
    key: str
    threshold_unit: str
    buckets: Mapping[str, Decimal]
    count: Optional[int]
    sum: Optional[Decimal]

    type = MetricTypes.HISTOGRAM

    def update(
        self,
        buckets: Mapping[str, Decimal],
        count: Optional[int] = None,
        sum: Optional[Decimal] = None,
    ):
        self.buckets = buckets
        self.count = count
        self.sum = sum

    def to_serializable_dict(self) -> Mapping[str, Any]:
        return {
            "__type": self.type.name,
            "current": {k: str(v) for k, v in self.buckets.items()},
            "threshold_unit": self.threshold_unit,
            "count": self.count,
            "sum": str(self.sum) if self.sum is not None else None,
        }


class FrontendServerMode(enum.StrEnum):
    WILDCARD_DOMAIN = "wildcard"
    PORT = "port"
    TRAEFIK = "traefik"


LAST_USED_MARKER_SOCKET_NAME: Final[str] = "last-used-time-marker.socket"



from __future__ import annotations

from .composer import DependencyInput, DependencyResources, WorkerDependencyComposer

__all__ = [
    "DependencyInput",
    "DependencyResources",
    "WorkerDependencyComposer",
]



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path

from ai.backend.common.dependencies import DependencyComposer, DependencyStack

from ...config import ServerConfig
from .config import ConfigInput, ConfigProvider


@dataclass
class BootstrapInput:
    """Input for bootstrap layer."""

    config_path: Path | None


@dataclass
class BootstrapResources:
    """Bootstrap resources for app proxy worker."""

    config: ServerConfig


class BootstrapComposer(DependencyComposer[BootstrapInput, BootstrapResources]):
    """Composer for bootstrap layer dependencies."""

    @property
    def stage_name(self) -> str:
        return "bootstrap"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: BootstrapInput,
    ) -> AsyncIterator[BootstrapResources]:
        """Compose bootstrap dependencies.

        Args:
            stack: The dependency stack to use for composition
            setup_input: Bootstrap input containing config path

        Yields:
            BootstrapResources containing loaded config
        """
        # Load configuration
        config = await stack.enter_dependency(
            ConfigProvider(),
            ConfigInput(config_path=setup_input.config_path),
        )

        yield BootstrapResources(config=config)



import logging

import aiohttp
from aiohttp.client_exceptions import ClientConnectorError

from ai.backend.appproxy.common.exceptions import (
    CoordinatorConnectionError,
    InternalServerError,
    ObjectNotFound,
    ServerMisconfiguredError,
    WorkerRegistrationError,
)
from ai.backend.appproxy.common.types import (
    FrontendMode,
    FrontendServerMode,
    SerializableCircuit,
    Slot,
)
from ai.backend.appproxy.common.types import SerializableToken as Token
from ai.backend.common.clients.http_client.client_pool import ClientKey
from ai.backend.logging import BraceStyleAdapter

from .types import (
    LAST_USED_MARKER_SOCKET_NAME,
    Circuit,
    RootContext,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


def _get_coordinator_session(root_ctx: RootContext) -> aiohttp.ClientSession:
    """Get a pooled ClientSession for coordinator requests."""
    coordinator_endpoint = root_ctx.local_config.proxy_worker.coordinator_endpoint
    client_key = ClientKey(
        endpoint=str(coordinator_endpoint),
        domain="coordinator",
    )
    return root_ctx.http_client_pool.load_client_session(client_key)


def _get_request_headers(root_ctx: RootContext, request_id: str) -> dict[str, str]:
    """Get request-specific headers for coordinator API calls."""
    return {
        "X-BackendAI-RequestID": request_id,
        "X-BackendAI-Token": root_ctx.local_config.secrets.api_secret,
    }


async def get_circuit_info(root_ctx: RootContext, request_id: str, circuit_id: str) -> Circuit:
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    try:
        async with sess.get(f"/api/circuit/{circuit_id}", headers=headers) as resp:
            resp.raise_for_status()
            body = await resp.json()
            return Circuit.from_serialized_circuit(SerializableCircuit(**body))
    except aiohttp.ClientResponseError as e:
        if e.code == 404:
            raise ObjectNotFound(object_name="worker:circuit")
        else:
            log.exception("error while communicating with coordinator:")
            raise InternalServerError from e


async def list_worker_circuits(root_ctx: RootContext, request_id: str) -> list[Circuit]:
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    try:
        async with sess.get(f"/api/worker/{root_ctx.worker_id}/circuits", headers=headers) as resp:
            resp.raise_for_status()
            body = await resp.json()
            return [
                Circuit.from_serialized_circuit(SerializableCircuit(**c)) for c in body["circuits"]
            ]
    except aiohttp.ClientResponseError as e:
        if e.code == 404:
            raise ObjectNotFound(object_name="worker:worker")
        else:
            log.exception("error while communicating with coordinator:")
            raise InternalServerError from e


async def destroy_circuit(root_ctx: RootContext, request_id: str, circuit_id: str) -> None:
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    try:
        async with sess.delete(f"/api/circuit/{circuit_id}", headers=headers) as resp:
            resp.raise_for_status()
    except aiohttp.ClientResponseError as e:
        if e.code == 404:
            raise ObjectNotFound(object_name="worker:circuit")
        else:
            log.exception("error while communicating with coordinator:")
            raise InternalServerError from e


async def register_worker(root_ctx: RootContext, request_id: str) -> list[Slot]:
    local_config = root_ctx.local_config
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)

    frontend_mode: FrontendServerMode | FrontendMode
    if local_config.proxy_worker.frontend_mode == FrontendServerMode.TRAEFIK:
        if not local_config.proxy_worker.traefik:
            raise ServerMisconfiguredError("worker:proxy-worker.traefik")
        frontend_mode = local_config.proxy_worker.traefik.frontend_mode
    else:
        frontend_mode = local_config.proxy_worker.frontend_mode
    body: dict = {
        "authority": local_config.proxy_worker.authority,
        "frontend_mode": frontend_mode,
        "protocol": local_config.proxy_worker.protocol,
        "hostname": local_config.proxy_worker.api_advertised_addr.host
        if local_config.proxy_worker.api_advertised_addr
        else local_config.proxy_worker.api_bind_addr.host,
        "tls_listen": local_config.proxy_worker.tls_listen,
        "tls_advertised": local_config.proxy_worker.tls_advertised,
        "api_port": local_config.proxy_worker.api_advertised_addr.port
        if local_config.proxy_worker.api_advertised_addr
        else local_config.proxy_worker.api_bind_addr.port,
        "accepted_traffics": local_config.proxy_worker.accepted_traffics,
        "filtered_apps_only": local_config.proxy_worker.filtered_apps_only,
        "app_filters": local_config.proxy_worker.app_filters,
        "traefik_last_used_marker_path": (
            local_config.proxy_worker.traefik.last_used_time_marker_directory
            / LAST_USED_MARKER_SOCKET_NAME
        ).as_posix()
        if local_config.proxy_worker.traefik
        else None,
    }

    match local_config.proxy_worker.frontend_mode:
        case FrontendServerMode.PORT:
            if not local_config.proxy_worker.port_proxy:
                raise ServerMisconfiguredError("worker:proxy-worker.port-proxy")
            body["port_range"] = (
                local_config.proxy_worker.port_proxy.advertised_port_range
                or local_config.proxy_worker.port_proxy.bind_port_range
            )
        case FrontendServerMode.WILDCARD_DOMAIN:
            if not local_config.proxy_worker.wildcard_domain:
                raise ServerMisconfiguredError("worker:proxy-worker.wildcard-domain")
            body["wildcard_domain"] = local_config.proxy_worker.wildcard_domain.domain
            body["wildcard_traffic_port"] = (
                local_config.proxy_worker.wildcard_domain.advertised_port
                or local_config.proxy_worker.wildcard_domain.bind_addr.port
            )
        case FrontendServerMode.TRAEFIK:
            if not local_config.proxy_worker.traefik:
                raise ServerMisconfiguredError("worker:proxy-worker.traefik")
            match local_config.proxy_worker.traefik.frontend_mode:
                case FrontendMode.PORT:
                    if not local_config.proxy_worker.traefik.port_proxy:
                        raise ServerMisconfiguredError("worker:proxy-worker.traefik.port-proxy")
                    body["port_range"] = local_config.proxy_worker.traefik.port_proxy.port_range
                case FrontendMode.WILDCARD_DOMAIN:
                    if not local_config.proxy_worker.traefik.wildcard_domain:
                        raise ServerMisconfiguredError(
                            "worker:proxy-worker.traefik.wildcard-domain"
                        )
                    body["wildcard_domain"] = (
                        local_config.proxy_worker.traefik.wildcard_domain.domain
                    )
                    body["wildcard_traffic_port"] = (
                        local_config.proxy_worker.traefik.wildcard_domain.advertised_port
                    )
    try:
        async with sess.put("/api/worker", json=body, headers=headers) as resp:
            resp.raise_for_status()
            body = await resp.json()
            root_ctx.worker_id = body["id"]
            log.debug(
                "Joined to coordinator {}",
                root_ctx.local_config.proxy_worker.coordinator_endpoint,
            )
            return body["slots"]
    except aiohttp.ClientResponseError as e:
        log.exception("")
        if e.status == 400:
            log.warning("Error from coordinator: {}", e.message)
        raise WorkerRegistrationError(
            extra_data={
                "authority": local_config.proxy_worker.authority,
                "coordinator": str(root_ctx.local_config.proxy_worker.coordinator_endpoint),
            }
        ) from e
    except ClientConnectorError as e:
        raise CoordinatorConnectionError(
            extra_data={
                "authority": local_config.proxy_worker.authority,
                "coordinator": str(root_ctx.local_config.proxy_worker.coordinator_endpoint),
            }
        ) from e


async def deregister_worker(root_ctx: RootContext, request_id: str) -> None:
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    async with sess.delete(f"/api/worker/{root_ctx.worker_id}", headers=headers) as resp:
        resp.raise_for_status()


async def ping_worker(root_ctx: RootContext, request_id: str) -> None:
    worker_config = root_ctx.local_config.proxy_worker
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    try:
        async with sess.patch(f"/api/worker/{root_ctx.worker_id}", headers=headers) as resp:
            resp.raise_for_status()
    except ClientConnectorError as e:
        raise CoordinatorConnectionError(
            extra_data={
                "authority": worker_config.authority,
                "coordinator": str(worker_config.coordinator_endpoint),
            }
        ) from e


async def get_token_data(root_ctx: RootContext, request_id: str, token_id: str) -> Token:
    sess = _get_coordinator_session(root_ctx)
    headers = _get_request_headers(root_ctx, request_id)
    async with sess.get(f"/api/worker/token/{token_id}", headers=headers) as resp:
        resp.raise_for_status()
        return Token(**(await resp.json()))



import asyncio
import logging
from asyncio import subprocess

from yarl import URL

from ai.backend.appproxy.common.exceptions import ServerMisconfiguredError
from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.appproxy.worker.proxy.backend.h2 import BackendConfig, H2Backend
from ai.backend.appproxy.worker.types import Circuit, SubdomainFrontendInfo
from ai.backend.logging import BraceStyleAdapter

from .base import H2Frontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class SubdomainFrontend(H2Frontend[str]):
    process: subprocess.Process | None
    log_monitor_tasks: list[asyncio.Task]
    proc_monitor_task: asyncio.Task | None

    api_port: int | None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.process = None
        self.log_monitor_tasks = []
        self.proc_monitor_task = None

    async def start(self) -> None:
        self.api_port = self.api_port_pool.pop()
        worker_config = self.root_context.local_config.proxy_worker
        wildcard_domain_config = worker_config.wildcard_domain
        if not wildcard_domain_config:
            raise ServerMisconfiguredError("worker:proxy-worker.wildcard-domain")

        service_addr = wildcard_domain_config.bind_addr
        nghttpx_args = ["-s", f"--frontend=127.0.0.1,{self.api_port};api"]
        if worker_config.tls_listen:
            nghttpx_args += [
                f"--frontend={service_addr.host},{service_addr.port}",
                str(worker_config.tls_cert),
                str(worker_config.tls_privkey),
            ]
        else:
            nghttpx_args.append(f"--frontend={service_addr.host},{service_addr.port};no-tls")

        if not self.root_context.local_config.proxy_worker.http2:
            raise ServerMisconfiguredError("worker:proxy-worker.http2")
        proc = await subprocess.create_subprocess_exec(
            self.root_context.local_config.proxy_worker.http2.nghttpx_path,
            *nghttpx_args,
        )
        assert proc.stdout and proc.stderr
        self.process = proc
        self.log_monitor_tasks.append(
            asyncio.create_task(self._log_monitor_task(proc.stdout, "stdout"))
        )
        self.log_monitor_tasks.append(
            asyncio.create_task(self._log_monitor_task(proc.stderr, "stderr"))
        )
        self.proc_monitor_task = asyncio.create_task(self._proc_monitor_task(proc))
        log.info("accepting proxy requests at {}:{}", service_addr.host, service_addr.port)

    async def stop(self) -> None:
        if not self.proc_monitor_task:
            return
        self.proc_monitor_task.cancel()
        await self.proc_monitor_task
        if not self.process:
            return
        self.process.terminate()
        await self.process.wait()
        try:
            for task in self.log_monitor_tasks:
                task.cancel()
                await task
        finally:
            if self.api_port:
                self.api_port_pool.add(self.api_port)

    async def initialize_backend(self, circuit: Circuit, routes: list[RouteInfo]) -> H2Backend:
        backend = H2Backend(URL(f"http://localhost:{self.api_port}"), self.root_context, circuit)
        return await self.update_backend(backend, routes)

    async def update_backend(self, backend: H2Backend, routes: list[RouteInfo]) -> H2Backend:
        backend_configs = [BackendConfig(r.current_kernel_host, r.kernel_port) for r in routes]
        await backend.update_config(backend_configs)
        return backend

    async def terminate_backend(self, backend: H2Backend) -> None:
        await backend.update_config([])

    def get_circuit_key(self, circuit: Circuit) -> str:
        assert isinstance(circuit.frontend, SubdomainFrontendInfo)
        return circuit.frontend.subdomain.lower()



import asyncio
import logging
from collections import defaultdict
from decimal import Decimal
from typing import Any, Final
from uuid import UUID

from prometheus_client.parser import text_string_to_metric_families

from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.common.clients.http_client.client_pool import ClientKey, ClientPool
from ai.backend.common.types import MetricKey, ModelServiceStatus, RuntimeVariant
from ai.backend.logging import BraceStyleAdapter

from .types import (
    Circuit,
    HistogramMeasurement,
    HistogramMetric,
    InferenceAppInfo,
    InferenceMeasurement,
    Measurement,
    Metric,
    MetricTypes,
    MovingStatistics,
    RootContext,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

CACHE_LIFESPAN: Final[int] = 120


def add_matrices(*ms: tuple[float | int | Decimal, ...]) -> tuple[Decimal, ...]:
    if len(ms) == 0:
        return tuple()
    matrix_size = len(ms[0])
    for m in ms:
        assert len(m) == matrix_size, "Inconsistent matrix size"
    base = [Decimal()] * matrix_size
    for m in ms:
        for i in range(matrix_size):
            base[i] += Decimal(m[i])
    return tuple(base)


async def gather_prometheus_inference_measures(
    client_pool: ClientPool,
    routes: list[RouteInfo],
    request_path: str = "/metrics",
) -> list[InferenceMeasurement]:
    histogram_metrics_labels: dict[str, tuple[str, ...]] = {}  # [metric name: (bucket label, ...)]

    histogram_bucket_metrics: defaultdict[str, dict[UUID, tuple[Decimal, ...]]] = defaultdict(
        dict
    )  # [metric name: [route ID: (bucket, ...)]]
    histogram_count_metrics: defaultdict[str, dict[UUID, int]] = defaultdict(
        dict
    )  # [metric name: [route ID: value]]
    histogram_sum_metrics: defaultdict[str, dict[UUID, Decimal]] = defaultdict(
        dict
    )  # [metric name: [route ID: value]]
    gauge_metrics: defaultdict[str, defaultdict[UUID, Decimal]] = defaultdict(
        lambda: defaultdict(Decimal)
    )  # [metric name: [route ID: value]]
    counter_metrics: defaultdict[str, defaultdict[UUID, Decimal]] = defaultdict(
        lambda: defaultdict(Decimal)
    )  # [metric name: [route ID: value]]

    # Use ClientPool to reuse sessions per kernel host:port
    for route in routes:
        if not route.route_id:
            continue

        # Skip unhealthy routes to avoid connection errors
        if route.health_status and route.health_status != ModelServiceStatus.HEALTHY:
            log.debug("Skipping metrics collection for unhealthy route {}", route.route_id)
            continue

        # Create a client key for each kernel endpoint
        client_key = ClientKey(
            endpoint=f"http://{route.current_kernel_host}:{route.kernel_port}",
            domain=str(route.route_id),
        )
        client_session = client_pool.load_client_session(client_key)

        async with client_session.get(request_path) as resp:
            resp.raise_for_status()
            metrics_text = await resp.text()
            metric_families = text_string_to_metric_families(metrics_text)
            for metric_family in metric_families:
                metric_name = metric_family.name
                match metric_family.type:
                    case "histogram":
                        labels: list[str] = []
                        values: list[Decimal] = []
                        for sample in metric_family.samples:
                            if sample.name.endswith("_bucket"):
                                labels.append(sample.labels["le"])
                                values.append(Decimal(sample.value))
                            elif sample.name.endswith("_count"):
                                histogram_count_metrics[metric_name][route.route_id] = int(
                                    sample.value
                                )
                            elif sample.name.endswith("_sum"):
                                histogram_sum_metrics[metric_name][route.route_id] = Decimal(
                                    sample.value
                                )
                        histogram_metrics_labels[metric_name] = tuple(labels)
                        histogram_bucket_metrics[metric_name][route.route_id] = tuple(values)
                    case "gauge":
                        try:
                            value = metric_family.samples[0].value
                        except IndexError:
                            continue
                        gauge_metrics[metric_name][route.route_id] = Decimal(value)
                    case "counter":
                        try:
                            value = metric_family.samples[0].value
                        except IndexError:
                            continue
                        counter_metrics[metric_name][route.route_id] = Decimal(value)

    measures: list[InferenceMeasurement] = []
    for metric_name, per_route_histogram_metrics in histogram_bucket_metrics.items():
        aggregated_buckets = add_matrices(*per_route_histogram_metrics.values())
        aggregated_count = sum([
            d for d in histogram_count_metrics[metric_name].values() if d is not None
        ])
        aggregated_sum = sum([
            d for d in histogram_sum_metrics[metric_name].values() if d is not None
        ]) or Decimal("0")  # sum returns Literal['0'] if empty

        aggregated_histogram_measurement = HistogramMeasurement(
            buckets={
                label: Decimal(value)
                for label, value in zip(histogram_metrics_labels[metric_name], aggregated_buckets)
            },
            count=aggregated_count,
            sum=aggregated_sum,
        )
        per_replica_histogram_measurements = {
            route_id: HistogramMeasurement(
                buckets={
                    label: Decimal(value)
                    for label, value in zip(histogram_metrics_labels[metric_name], values)
                },
                count=count,
                sum=sum,
            )
            for route_id, values, count, sum in zip(
                per_route_histogram_metrics.keys(),
                per_route_histogram_metrics.values(),
                [
                    histogram_count_metrics[metric_name].get(route_id)
                    for route_id in per_route_histogram_metrics.keys()
                ],
                [
                    histogram_sum_metrics[metric_name].get(route_id)
                    for route_id in per_route_histogram_metrics.keys()
                ],
            )
        }
        measures.append(
            InferenceMeasurement(
                key=MetricKey(metric_name),
                type=MetricTypes.HISTOGRAM,
                per_app=aggregated_histogram_measurement,
                per_replica=per_replica_histogram_measurements,
            )
        )
    for metric_name, per_route_gauge_metrics in gauge_metrics.items():
        aggregated_gauge_metric = Measurement(Decimal(sum(per_route_gauge_metrics.values())))
        per_replica_gauge_measurements = {
            route_id: Measurement(value) for route_id, value in per_route_gauge_metrics.items()
        }
        measures.append(
            InferenceMeasurement(
                key=MetricKey(metric_name),
                type=MetricTypes.GAUGE,
                per_app=aggregated_gauge_metric,
                per_replica=per_replica_gauge_measurements,
            )
        )
    for metric_name, per_route_gauge_metrics in counter_metrics.items():
        aggregated_counter_metric = Measurement(Decimal(sum(per_route_gauge_metrics.values())))
        per_replica_gauge_measurements = {
            route_id: Measurement(value) for route_id, value in per_route_gauge_metrics.items()
        }
        measures.append(
            InferenceMeasurement(
                key=MetricKey(metric_name),
                type=MetricTypes.GAUGE,
                per_app=aggregated_counter_metric,
                per_replica=per_replica_gauge_measurements,
            )
        )
    return measures


async def gather_inference_measures(
    client_pool: ClientPool, circuit: Circuit
) -> list[InferenceMeasurement] | None:
    assert isinstance(circuit.app_info, InferenceAppInfo)
    measures: list[InferenceMeasurement] = []

    match circuit.app_info.runtime_variant:
        case RuntimeVariant.VLLM:
            raw_measures = await gather_prometheus_inference_measures(
                client_pool, circuit.route_info
            )

            for measure in raw_measures:
                if not measure.key.startswith("vllm:"):
                    continue
                measures.append(
                    InferenceMeasurement(
                        key=MetricKey(measure.key.replace("vllm:", "vllm_")),
                        type=MetricTypes.GAUGE,
                        per_app=measure.per_app,
                        per_replica=measure.per_replica,
                    )
                )
            return measures
        case RuntimeVariant.HUGGINGFACE_TGI:
            return await gather_prometheus_inference_measures(client_pool, circuit.route_info)
        case RuntimeVariant.SGLANG:
            raw_measures = await gather_prometheus_inference_measures(
                client_pool, circuit.route_info
            )

            for measure in raw_measures:
                if not measure.key.startswith("sglang:"):
                    continue
                measures.append(
                    InferenceMeasurement(
                        key=MetricKey(measure.key.replace("sglang:", "sglang_")),
                        type=MetricTypes.GAUGE,
                        per_app=measure.per_app,
                        per_replica=measure.per_replica,
                    )
                )
            return measures
        case RuntimeVariant.MODULAR_MAX:
            raw_measures = await gather_prometheus_inference_measures(
                client_pool, circuit.route_info
            )

            for measure in raw_measures:
                if not measure.key.startswith("max:"):
                    continue
                measures.append(
                    InferenceMeasurement(
                        key=MetricKey(measure.key.replace("max:", "max_")),
                        type=MetricTypes.GAUGE,
                        per_app=measure.per_app,
                        per_replica=measure.per_replica,
                    )
                )
            return measures
        case _:
            return None


async def collect_inference_metric(root_ctx: RootContext, interval: float) -> None:
    try:
        inference_circuits = [
            circuit
            for circuit in root_ctx.proxy_frontend.circuits.values()
            if isinstance(circuit.app_info, InferenceAppInfo)
        ]

        # Here we use asyncio.gather() instead of aiotools.TaskGroup
        # to keep methods of other plugins running when a plugin raises an error
        # instead of cancelling them.
        _tasks: list[asyncio.Task[list[InferenceMeasurement] | None]] = [
            asyncio.create_task(gather_inference_measures(root_ctx.http_client_pool, c))
            for c in inference_circuits
        ]

        results = await asyncio.gather(*_tasks, return_exceptions=True)
        for circuit, result in zip(inference_circuits, results):
            if not result:  # Unsupported runtime variant (custom, Triton, ...)
                continue
            if isinstance(result, BaseException):
                log.error(
                    "collect_inference_metric(): gather_inference_measures() error", exc_info=result
                )
                continue
            for inference_measure in result:
                metric_key = inference_measure.key
                # update per-app metric
                if metric_key not in circuit._app_inference_metrics:
                    match inference_measure.per_app:
                        case Measurement():
                            circuit._app_inference_metrics[metric_key] = Metric(
                                metric_key,
                                inference_measure.type,
                                current=inference_measure.per_app.value,
                                capacity=inference_measure.per_app.capacity,
                                unit_hint=inference_measure.unit_hint,
                                stats=MovingStatistics(inference_measure.per_app.value),
                                stats_filter=frozenset(inference_measure.stats_filter),
                            )
                        case HistogramMeasurement():
                            circuit._app_inference_metrics[metric_key] = HistogramMetric(
                                metric_key,
                                "le",  # FIXME: customize
                                buckets=inference_measure.per_app.buckets,
                                count=inference_measure.per_app.count,
                                sum=inference_measure.per_app.sum,
                            )
                else:
                    existing_metric = circuit._app_inference_metrics[metric_key]
                    match inference_measure.per_app:
                        case Measurement():
                            if not isinstance(existing_metric, Metric):
                                raise ValueError(
                                    f"Unexpected metric type: {type(existing_metric)}, expected Metric"
                                )
                            existing_metric.update(inference_measure.per_app)
                        case HistogramMeasurement():
                            if not isinstance(existing_metric, HistogramMetric):
                                raise ValueError(
                                    f"Unexpected metric type: {type(existing_metric)}, expected HistogramMetric"
                                )
                            existing_metric.update(
                                inference_measure.per_app.buckets,
                                count=inference_measure.per_app.count,
                                sum=inference_measure.per_app.sum,
                            )
                # update per-replica metric
                for route_id, replica_inference_metrics in inference_measure.per_replica.items():
                    if metric_key not in circuit._replica_inference_metrics:
                        circuit._replica_inference_metrics[metric_key] = {}
                    if route_id not in circuit._replica_inference_metrics[metric_key]:
                        match replica_inference_metrics:
                            case Measurement():
                                circuit._replica_inference_metrics[metric_key][route_id] = Metric(
                                    metric_key,
                                    inference_measure.type,
                                    current=replica_inference_metrics.value,
                                    capacity=replica_inference_metrics.capacity,
                                    unit_hint=inference_measure.unit_hint,
                                    stats=MovingStatistics(replica_inference_metrics.value),
                                    stats_filter=frozenset(inference_measure.stats_filter),
                                )
                            case HistogramMeasurement():
                                circuit._replica_inference_metrics[metric_key][route_id] = (
                                    HistogramMetric(
                                        metric_key,
                                        "le",  # FIXME: customize
                                        buckets=replica_inference_metrics.buckets,
                                        count=replica_inference_metrics.count,
                                        sum=replica_inference_metrics.sum,
                                    )
                                )
                    else:
                        existing_metric = circuit._replica_inference_metrics[metric_key][route_id]
                        match replica_inference_metrics:
                            case Measurement():
                                if not isinstance(existing_metric, Metric):
                                    raise ValueError(
                                        f"Unexpected metric type: {type(existing_metric)}, expected Metric"
                                    )
                                existing_metric.update(replica_inference_metrics)
                            case HistogramMeasurement():
                                if not isinstance(existing_metric, HistogramMetric):
                                    raise ValueError(
                                        f"Unexpected metric type: {type(existing_metric)}, expected HistogramMetric"
                                    )
                                existing_metric.update(
                                    replica_inference_metrics.buckets,
                                    count=replica_inference_metrics.count,
                                    sum=replica_inference_metrics.sum,
                                )

        # push to the Redis server
        app_metrics_updates = {
            circuit.endpoint_id: {
                key: obj.to_serializable_dict()
                for key, obj in circuit._app_inference_metrics.items()
            }
            for circuit in inference_circuits
            if circuit._app_inference_metrics and circuit.endpoint_id
        }
        replica_metrics_updates: dict[tuple[UUID, UUID], Any] = {}

        for circuit in inference_circuits:
            if not circuit._replica_inference_metrics or not circuit.endpoint_id:
                continue
            for key, metric_pair in circuit._replica_inference_metrics.items():
                for route_id, obj in metric_pair.items():
                    if (circuit.endpoint_id, route_id) not in replica_metrics_updates:
                        replica_metrics_updates[(circuit.endpoint_id, route_id)] = {}
                    replica_metrics_updates[(circuit.endpoint_id, route_id)][key] = (
                        obj.to_serializable_dict()
                    )

        if root_ctx.local_config.debug.log_stats:
            log.debug(
                "stats: app_updates: {}",
                app_metrics_updates,
            )
            log.debug(
                "stats: replica_updates: {}",
                replica_metrics_updates,
            )

        await root_ctx.valkey_stat.store_inference_metrics(
            app_metrics_updates,
            replica_metrics_updates,
            CACHE_LIFESPAN,
        )
    except Exception:
        log.exception("collect_inference_metrics(): error while collecting metric:")
        raise



import asyncio
import logging
from asyncio import subprocess

from yarl import URL

from ai.backend.appproxy.common.exceptions import ServerMisconfiguredError
from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.appproxy.worker.proxy.backend.h2 import BackendConfig, H2Backend
from ai.backend.appproxy.worker.types import Circuit, PortFrontendInfo
from ai.backend.logging import BraceStyleAdapter

from .base import H2Frontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class PortFrontend(H2Frontend[int]):
    processes: list[subprocess.Process]
    healthy: bool
    proc_monitor_tasks: list[asyncio.Task] = []
    log_monitor_tasks: list[asyncio.Task] = []

    api_ports: dict[int, int]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.api_endpoint = None
        self.processes = []
        self.api_ports = {}

    async def start(self) -> None:
        worker_config = self.root_context.local_config.proxy_worker
        port_proxy_config = worker_config.port_proxy
        h2_config = worker_config.http2
        if not port_proxy_config:
            raise ServerMisconfiguredError("worker:proxy-worker.port-proxy")
        if not h2_config:
            raise ServerMisconfiguredError("worker:proxy-worker.http2")

        port_start, port_end = port_proxy_config.bind_port_range
        for listen_port in range(port_start, port_end + 1):
            api_port = self.api_port_pool.pop()
            nghttpx_args = ["-s", f"--frontend=127.0.0.1,{api_port};api"]
            if worker_config.tls_listen:
                nghttpx_args += [
                    f"--frontend={port_proxy_config.bind_host},{listen_port}",
                    str(worker_config.tls_cert),
                    str(worker_config.tls_privkey),
                ]
            else:
                nghttpx_args.append(f"--frontend=*,{listen_port};no-tls")

            proc = await subprocess.create_subprocess_exec(
                h2_config.nghttpx_path,
                *nghttpx_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            assert proc.stdout and proc.stderr
            self.processes.append(proc)
            self.log_monitor_tasks.append(
                asyncio.create_task(self._log_monitor_task(proc.stdout, f"stdout #{listen_port}"))
            )
            self.log_monitor_tasks.append(
                asyncio.create_task(self._log_monitor_task(proc.stderr, f"stderr #{listen_port}"))
            )
            self.proc_monitor_tasks.append(asyncio.create_task(self._proc_monitor_task(proc)))
            self.api_ports[listen_port] = api_port
            log.info("started nghttpx server at {}:{}", port_proxy_config.bind_host, listen_port)

    async def stop(self) -> None:
        for task in self.proc_monitor_tasks:
            task.cancel()
            await task
        for process in self.processes:
            process.terminate()
            await process.wait()
        try:
            for task in self.log_monitor_tasks:
                task.cancel()
                await task
        finally:
            self.api_ports = {}

    async def initialize_backend(self, circuit: Circuit, routes: list[RouteInfo]) -> H2Backend:
        api_port = self.api_ports[self.get_circuit_key(circuit)]
        backend = H2Backend(URL(f"http://localhost:{api_port}"), self.root_context, circuit)
        return await self.update_backend(backend, routes)

    async def update_backend(self, backend: H2Backend, routes: list[RouteInfo]) -> H2Backend:
        backend_configs = [BackendConfig(r.current_kernel_host, r.kernel_port) for r in routes]
        await backend.update_config(backend_configs)
        return backend

    async def terminate_backend(self, backend: H2Backend) -> None:
        await backend.update_config([])

    def get_circuit_key(self, circuit: Circuit) -> int:
        assert isinstance(circuit.frontend, PortFrontendInfo)
        return circuit.frontend.port



from __future__ import annotations

import asyncio
import logging
import random
import time
from contextlib import asynccontextmanager
from functools import partial
from typing import AsyncIterator, Final, override

import aiohttp
import aiotools
from aiohttp import ClientConnectorError, web
from multidict import CIMultiDict
from yarl import URL

from ai.backend.appproxy.common.exceptions import ContainerConnectionRefused, WorkerNotAvailable
from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.common.clients.http_client.client_pool import (
    ClientKey,
    ClientPool,
    tcp_client_session_factory,
)
from ai.backend.logging import BraceStyleAdapter

from .base import BaseBackend, HttpRequest

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

HOP_ONLY_HEADERS: Final[CIMultiDict[int]] = CIMultiDict([
    ("Connection", 1),
    ("Keep-Alive", 1),
    ("Proxy-Authenticate", 1),
    ("Proxy-Authorization", 1),
    ("TE", 1),
    ("Trailers", 1),
    ("Transfer-Encoding", 1),
    ("Upgrade", 1),
])


class HTTPBackend(BaseBackend):
    routes: list[RouteInfo]

    def __init__(self, routes: list[RouteInfo], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.routes = routes
        client_timeout = aiohttp.ClientTimeout(
            total=None,
            connect=10.0,
            sock_connect=10.0,
            sock_read=None,
        )
        cleanup_interval = self.root_context.local_config.proxy_worker.client_pool_cleanup_interval
        self.client_pool = ClientPool(
            partial(
                tcp_client_session_factory,
                timeout=client_timeout,
                auto_decompress=False,  # transparently pass the response body
            ),
            cleanup_interval_seconds=cleanup_interval,
        )

    @override
    async def close(self) -> None:
        await self.client_pool.close()

    @property
    def selected_route(self) -> RouteInfo:
        if len(self.routes) == 0:
            raise WorkerNotAvailable
        elif len(self.routes) == 1:
            selected_route = self.routes[0]
            if selected_route.traffic_ratio == 0:
                raise WorkerNotAvailable
        else:
            ratios: list[float] = [r.traffic_ratio for r in self.routes]
            selected_route = random.choices(self.routes, weights=ratios, k=1)[0]
        return selected_route

    def get_x_forwarded_proto(self, request: web.Request) -> str:
        use_tls = (
            self.root_context.local_config.proxy_worker.tls_advertised
            or self.root_context.local_config.proxy_worker.tls_listen
        )
        return request.headers.get("x-forwarded-proto") or ("https" if use_tls else "http")

    def get_x_forwarded_host(self, request: web.Request) -> str | None:
        return request.headers.get("x-forwarded-host") or request.headers.get("host")

    @asynccontextmanager
    async def request_http(
        self, route: RouteInfo, request: HttpRequest
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        metrics = self.root_context.metrics
        remote = f"{route.current_kernel_host}:{route.kernel_port}"
        metrics.proxy.observe_upstream_http_request(remote=remote, total_bytes_size=0)
        client_key = ClientKey(
            endpoint=f"http://{route.current_kernel_host}:{route.kernel_port}",
            domain=str(route.route_id),
        )
        client_session = self.client_pool.load_client_session(client_key)
        async with client_session.request(
            request.method,
            request.path,
            headers=request.headers,
            allow_redirects=False,
            data=request.body,
            auto_decompress=False,  # transparently pass the response body
            compress=None,  # transparently pass the request body
        ) as response:
            metrics.proxy.observe_upstream_http_response(remote=remote, total_bytes_size=0)
            yield response

    @asynccontextmanager
    async def connect_websocket(
        self, route: RouteInfo, request: web.Request, protocols: list[str] = []
    ) -> AsyncIterator[aiohttp.ClientWebSocketResponse]:
        client_key = ClientKey(
            endpoint=f"http://{route.current_kernel_host}:{route.kernel_port}",
            domain=str(route.route_id),
        )
        client_session = self.client_pool.load_client_session(client_key)
        log.trace("connecting to {}", URL(client_key.endpoint).with_path(request.path))
        async with client_session.ws_connect(request.rel_url, protocols=protocols) as ws:
            yield ws

    async def proxy_http(self, frontend_request: web.Request) -> web.StreamResponse:
        protocol = self.get_x_forwarded_proto(frontend_request)
        host = self.get_x_forwarded_host(frontend_request)
        remote_host, remote_port = (
            frontend_request.transport.get_extra_info("peername")
            if frontend_request.transport
            else None,
            None,
        )
        backend_rqst_hdrs = {}
        # copy frontend request headers without hop-by-hop headers
        for key, value in frontend_request.headers.items():
            if key not in HOP_ONLY_HEADERS:
                backend_rqst_hdrs[key] = value
        # overwrite proxy-related headers
        backend_rqst_hdrs["x-forwarded-proto"] = protocol
        if self.circuit.app == "rstudio":
            backend_rqst_hdrs["x-rstudio-proto"] = protocol
        if host:
            backend_rqst_hdrs["forwarded"] = f"host={host};proto={protocol}"
            backend_rqst_hdrs["x-forwarded-host"] = host
            if self.circuit.app == "rstudio":
                backend_rqst_hdrs["x-rstudio-request"] = (
                    f"{protocol}://{host}{frontend_request.path or ''}"
                )
            split = host.split(":")
            if len(split) >= 2:
                backend_rqst_hdrs["x-forwarded-port"] = split[1]
            elif remote_port:
                backend_rqst_hdrs["x-forwarded-port"] = remote_port
        if remote_host:
            backend_rqst_hdrs["x-forwarded-for"] = f"{remote_host[0]}:{remote_host[1]}"
        if frontend_request.body_exists:
            backend_rqst_body = frontend_request.content.iter_any()
        else:
            backend_rqst_body = None
        backend_request = HttpRequest(
            frontend_request.method,
            frontend_request.rel_url,
            backend_rqst_hdrs,
            backend_rqst_body,
        )
        route = self.selected_route
        await self.mark_last_used_time(route)
        await self.increase_request_counter()
        log.trace(
            "proxying {} {} HTTP Request to {}:{}",
            frontend_request.method,
            frontend_request.rel_url,
            route.current_kernel_host,
            route.kernel_port,
        )
        try:
            async with self.request_http(route, backend_request) as backend_response:
                frontend_resp_hdrs = {}
                for key, value in backend_response.headers.items():
                    if key not in HOP_ONLY_HEADERS:
                        frontend_resp_hdrs[key] = value
                frontend_resp_hdrs["Access-Control-Allow-Origin"] = "*"
                frontend_response = web.StreamResponse(
                    status=backend_response.status,
                    reason=backend_response.reason,
                    headers=frontend_resp_hdrs,
                )
                if "Content-Length" not in backend_response.headers:
                    frontend_response.enable_chunked_encoding()
                await frontend_response.prepare(frontend_request)
                recv_len = 0
                try:
                    async for data in backend_response.content.iter_any():
                        recv_len += len(data)
                        await frontend_response.write(data)
                except aiohttp.ClientPayloadError as e:
                    log.exception(
                        "{!r} (recv-len: {}, content-length: {}, headers: {!r})",
                        e,
                        recv_len,
                        backend_response.content_length,
                        frontend_resp_hdrs,
                    )
                finally:
                    await frontend_response.write_eof()
                return frontend_response
        except aiohttp.ServerDisconnectedError as e:
            log.debug(
                "Backend container disconnected unexpectedly: {}:{}, method={}, path={}, error={}",
                route.current_kernel_host,
                route.kernel_port,
                frontend_request.method,
                frontend_request.rel_url,
                e,
            )
            raise
        except ConnectionResetError:
            raise asyncio.CancelledError()
        except aiohttp.ClientOSError as e:
            raise ContainerConnectionRefused from e
        except:
            log.exception("Unhandled exception while proxying HTTP request")
            raise

    async def proxy_ws(self, request: web.Request) -> web.WebSocketResponse:
        metrics = self.root_context.metrics
        stop_event = asyncio.Event()
        total_bytes = 0

        async def _proxy_task(
            left: web.WebSocketResponse | aiohttp.ClientWebSocketResponse,
            right: web.WebSocketResponse | aiohttp.ClientWebSocketResponse,
            tag="(unknown)",
        ) -> None:
            nonlocal total_bytes

            try:
                async for msg in left:
                    match msg.type:
                        case aiohttp.WSMsgType.TEXT:
                            total_bytes += len(msg.data)
                            metrics.proxy.observe_upstream_ws_traffic_chunk(
                                total_bytes_size=total_bytes
                            )
                            await right.send_str(msg.data)
                        case aiohttp.WSMsgType.BINARY:
                            total_bytes += len(msg.data)
                            metrics.proxy.observe_upstream_ws_traffic_chunk(
                                total_bytes_size=total_bytes
                            )
                            await right.send_bytes(msg.data)
                        case aiohttp.WSMsgType.PING:
                            await right.ping(msg.data)
                        case aiohttp.WSMsgType.PONG:
                            await right.pong(msg.data)
                        case aiohttp.WSMsgType.CLOSE:
                            log.debug("{}: websocket closed", tag)
                            await right.close(code=msg.data)
                        case aiohttp.WSMsgType.ERROR:
                            log.debug("{}: websocket closed with error", tag)
                            await right.close()
                        case _:
                            log.debug("{}: Unhandled message type {}", tag, msg.type)
            except ConnectionResetError:
                pass
            except Exception:
                log.exception("")
                raise
            finally:
                log.debug("setting stop event")
                stop_event.set()

        async def _last_access_marker_task(interval: float) -> None:
            await self.mark_last_used_time(route)

        route = self.selected_route
        log.trace(
            "Proxying {} {} WS Request to {}:{}",
            request.method,
            request.path or "/",
            route.current_kernel_host,
            route.kernel_port,
        )

        if "Sec-WebSocket-Protocol" in request.headers:
            protocols = list(request.headers["Sec-WebSocket-Protocol"].split(","))
        else:
            protocols = []

        downstream_ws = web.WebSocketResponse(protocols=protocols)
        await downstream_ws.prepare(request)

        metrics.proxy.observe_upstream_ws_connection_start()
        start = time.monotonic()

        marker_task = aiotools.create_timer(_last_access_marker_task, 1.5)
        await self.increase_request_counter()

        try:
            async with self.connect_websocket(route, request, protocols=protocols) as upstream_ws:
                try:
                    async with asyncio.TaskGroup() as group:
                        group.create_task(
                            _proxy_task(upstream_ws, downstream_ws, tag="(up -> down)")
                        )
                        group.create_task(
                            _proxy_task(downstream_ws, upstream_ws, tag="(down -> up)")
                        )
                        log.debug("created tasks, now waiting until one of two tasks end")
                        await stop_event.wait()
                finally:
                    marker_task.cancel()
                    await marker_task
                    if not downstream_ws.closed:
                        await downstream_ws.close()
                    if not upstream_ws.closed:
                        await upstream_ws.close()
            log.trace("websocket connection closed")
        except ClientConnectorError:
            log.trace("upstream connection closed")
            if not downstream_ws.closed:
                await downstream_ws.close()
        finally:
            end = time.monotonic()
            metrics.proxy.observe_upstream_ws_connection_end(
                duration=int(end - start), total_bytes_size=total_bytes
            )

        return downstream_ws



from typing import Iterable

import aiohttp_cors
from aiohttp import web

from ai.backend.appproxy.common.types import CORSOptions, FrontendMode, WebMiddleware

from .. import __version__
from ..types import RootContext


async def hello(request: web.Request) -> web.Response:
    """Health check endpoint with dependency connectivity status"""
    from ai.backend.common.dto.internal.health import HealthResponse, HealthStatus

    request["do_not_print_access_log"] = True

    root_ctx: RootContext = request.app["_root.context"]
    connectivity = await root_ctx.health_probe.get_connectivity_status()
    response = HealthResponse(
        status=HealthStatus.OK if connectivity.overall_healthy else HealthStatus.DEGRADED,
        version=__version__,
        component="appproxy-worker",
        connectivity=connectivity,
    )
    return web.json_response(response.model_dump_json())


async def status(request: web.Request) -> web.Response:
    """
    Returns health status of worker.
    """
    request["do_not_print_access_log"] = True

    root_ctx: RootContext = request.app["_root.context"]
    worker_config = root_ctx.local_config.proxy_worker
    if worker_config.frontend_mode == FrontendMode.WILDCARD_DOMAIN:
        available_slots = 0
    else:
        assert worker_config.port_proxy is not None
        available_slots = (
            worker_config.port_proxy.bind_port_range[1]
            - worker_config.port_proxy.bind_port_range[0]
            + 1
        )
    return web.json_response({
        "version": __version__,
        "authority": worker_config.authority,
        "app_mode": worker_config.frontend_mode,
        "protocol": worker_config.protocol,
        "occupied_slots": len(root_ctx.proxy_frontend.circuits),
        "available_slots": available_slots,
    })


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "health"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", hello))
    cors.add(add_route("GET", "/status", status))
    return app, []



import urllib.parse
from typing import Iterable
from uuid import UUID

import aiohttp
import jwt
import yarl
from aiohttp import web
from pydantic import AnyUrl, BaseModel
from tenacity import AsyncRetrying, TryAgain, retry_if_exception_type, wait_exponential
from tenacity.stop import stop_after_attempt

from ai.backend.appproxy.common.defs import PERMIT_COOKIE_NAME
from ai.backend.appproxy.common.exceptions import (
    InvalidAPIParameters,
    ServerMisconfiguredError,
)
from ai.backend.appproxy.common.types import (
    CORSOptions,
    FrontendMode,
    ProxyProtocol,
    PydanticResponse,
    WebMiddleware,
)
from ai.backend.appproxy.common.types import SerializableCircuit as Circuit
from ai.backend.appproxy.common.utils import calculate_permit_hash, pydantic_api_handler

from ..config import (
    PortProxyConfig,
    TraefikPortProxyConfig,
    TraefikWildcardDomainConfig,
    WildcardDomainConfig,
)
from ..coordinator_client import get_circuit_info
from ..types import FrontendServerMode, InteractiveAppInfo, RootContext


def generate_proxy_url(
    config: PortProxyConfig
    | WildcardDomainConfig
    | TraefikPortProxyConfig
    | TraefikWildcardDomainConfig,
    protocol: str,
    circuit: Circuit,
    redirect_path: str | None = None,
) -> str:
    # Generate base URL based on config type
    match config:
        case PortProxyConfig():
            base_url = f"{protocol}://{config.advertised_host or config.bind_host}:{circuit.port}"
        case TraefikPortProxyConfig():
            base_url = f"{protocol}://{config.advertised_host}:{circuit.port}"
        case WildcardDomainConfig():
            base_url = f"{protocol}://{circuit.subdomain}{config.domain}:{config.advertised_port or config.bind_addr.port}"
        case TraefikWildcardDomainConfig():
            base_url = f"{protocol}://{circuit.subdomain}{config.domain}:{config.advertised_port}"

    # Append redirect path if provided
    if redirect_path:
        # Ensure redirect path starts with /
        if not redirect_path.startswith("/"):
            redirect_path = f"/{redirect_path}"
        return f"{base_url}{redirect_path}"

    return base_url


async def ensure_traefik_route_set_up(traefik_api_port: int, circuit: Circuit) -> None:
    match circuit.protocol:
        case ProxyProtocol.TCP:
            proto = "tcp"
        case _:
            proto = "http"
    path = f"/api/{proto}/routers/{circuit.traefik_router_name}"

    base_url = yarl.URL("http://127.0.0.1").with_port(traefik_api_port)
    async for attempt in AsyncRetrying(
        wait=wait_exponential(multiplier=0.02, min=0.02, max=5.0),
        stop=stop_after_attempt(20),
        retry=retry_if_exception_type(TryAgain),
    ):
        with attempt:
            async with aiohttp.ClientSession(base_url=base_url) as sess:
                async with sess.get(path) as resp:
                    if resp.status == 404:
                        raise TryAgain


class ProxySetupRequestModel(BaseModel):
    token: str


class ProxySetupResponseModel(BaseModel):
    redirect: AnyUrl
    redirectURI: AnyUrl


@pydantic_api_handler(ProxySetupRequestModel)
async def setup(
    request: web.Request, params: ProxySetupRequestModel
) -> web.StreamResponse | PydanticResponse[ProxySetupResponseModel]:
    root_ctx: RootContext = request.app["_root.context"]
    jwt_body = jwt.decode(
        params.token, root_ctx.local_config.secrets.jwt_secret, algorithms=["HS256"]
    )
    requested_circuit_id = UUID(jwt_body["circuit"])

    config = root_ctx.local_config.proxy_worker
    port_config = config.port_proxy  # As a default fallback
    circuit = await get_circuit_info(root_ctx, request["request_id"], str(requested_circuit_id))

    match config.frontend_mode:
        case FrontendServerMode.TRAEFIK:
            if config.traefik is None:
                raise ServerMisconfiguredError("proxy_worker: Missing 'traefik' config section")
            match config.traefik.frontend_mode:
                case FrontendMode.PORT:
                    port_config = config.traefik.port_proxy
                    if port_config is None:
                        raise ServerMisconfiguredError(
                            "proxy_worker.traefik: Missing 'port_proxy' config section"
                        )
                case FrontendMode.WILDCARD_DOMAIN:
                    port_config = config.traefik.wildcard_domain
                    if port_config is None:
                        raise ServerMisconfiguredError(
                            "proxy_worker.traefik: Missing 'wildcard_domain' config section"
                        )
            await ensure_traefik_route_set_up(config.traefik.api_port, circuit)
        case FrontendServerMode.PORT:
            port_config = config.port_proxy
            if port_config is None:
                raise ServerMisconfiguredError(
                    "proxy_worker: Missing root-level 'port_proxy' config section"
                )
        case FrontendServerMode.WILDCARD_DOMAIN:
            port_config = config.wildcard_domain
            if port_config is None:
                raise ServerMisconfiguredError(
                    "proxy_worker: Missing root-level 'wildcard_domain' config section"
                )
        case _:
            raise ServerMisconfiguredError(
                f"proxy_worker: Invalid root-level 'frontend_mode': {config.frontend_mode}"
            )
    assert port_config is not None

    use_tls = config.tls_advertised or config.tls_listen
    if not isinstance(circuit.app_info, InteractiveAppInfo):
        raise InvalidAPIParameters("E20011: Not supported for inference apps")

    # Web browsers block redirect between cross-origins if Access-Control-Allow-Origin value is set to a concrete Origin instead of wildcard;
    # Hence we need to send "*" as allowed origin manually, instead of benefiting from aiohttp-cors
    cors_headers = {
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Headers": "*",
        "Access-Control-Expose-Headers": "*",
    }
    match circuit.protocol:
        case ProxyProtocol.HTTP:
            protocol = "https" if use_tls else "http"
            redirect_path = jwt_body.get("redirect", "")
            proxy_url = generate_proxy_url(port_config, protocol, circuit, redirect_path)
            response = web.HTTPPermanentRedirect(proxy_url, headers=cors_headers)
            cookie_domain = None
            if circuit.frontend_mode == FrontendMode.WILDCARD_DOMAIN:
                wildcard_info = config.wildcard_domain
                if not wildcard_info:
                    raise ServerMisconfiguredError("worker:proxy-worker.wildcard-domain")
                cookie_domain = wildcard_info.domain
            response.set_cookie(
                PERMIT_COOKIE_NAME,
                calculate_permit_hash(root_ctx.local_config.permit_hash, circuit.app_info.user_id),
                domain=cookie_domain,
            )
            return response
        case ProxyProtocol.TCP:
            protocol = "tcp"
            queryparams = {
                "directTCP": "true",
                "auth": params.token,
                "proto": protocol,
                "gateway": generate_proxy_url(port_config, protocol, circuit, redirect_path=None),
            }
            if jwt_body["redirect"]:
                return web.HTTPPermanentRedirect(
                    f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}",
                    headers=cors_headers,
                )
            else:
                return PydanticResponse(
                    ProxySetupResponseModel(
                        redirect=AnyUrl(
                            f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}"
                        ),
                        redirectURI=AnyUrl(
                            f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}"
                        ),
                    ),
                    headers=cors_headers,
                )
        case _:
            raise InvalidAPIParameters("E20002: Protocol not available as interactive app")


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "setup"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    add_route = app.router.add_route
    add_route("GET", "", setup)
    return app, []



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.defs import REDIS_LIVE_DB, REDIS_STATISTICS_DB, RedisRole
from ai.backend.common.dependencies import DependencyProvider
from ai.backend.common.health_checker import CID_REDIS_LIVE, CID_REDIS_STAT, ServiceHealthChecker
from ai.backend.common.health_checker.checkers.valkey import ValkeyHealthChecker
from ai.backend.common.types import RedisProfileTarget

from ...config import ServerConfig


@dataclass
class WorkerValkeyClients:
    """Valkey clients for app proxy worker."""

    valkey_live: ValkeyLiveClient
    valkey_stat: ValkeyStatClient


class RedisProvider(DependencyProvider[ServerConfig, WorkerValkeyClients]):
    """Provider for Redis configuration and Valkey clients."""

    @property
    def stage_name(self) -> str:
        return "redis"

    @asynccontextmanager
    async def provide(self, setup_input: ServerConfig) -> AsyncIterator[WorkerValkeyClients]:
        """Create and provide Valkey clients."""
        redis_profile_target = RedisProfileTarget.from_dict(setup_input.redis.to_dict())

        valkey_live = await ValkeyLiveClient.create(
            valkey_target=redis_profile_target.profile_target(RedisRole.LIVE).to_valkey_target(),
            db_id=REDIS_LIVE_DB,
            human_readable_name="appproxy-worker",
        )
        valkey_stat = await ValkeyStatClient.create(
            valkey_target=redis_profile_target.profile_target(
                RedisRole.STATISTICS
            ).to_valkey_target(),
            db_id=REDIS_STATISTICS_DB,
            human_readable_name="appproxy-worker",
        )

        try:
            yield WorkerValkeyClients(
                valkey_live=valkey_live,
                valkey_stat=valkey_stat,
            )
        finally:
            await valkey_live.close()
            await valkey_stat.close()

    def gen_health_checkers(self, resource: WorkerValkeyClients) -> ServiceHealthChecker:
        """
        Return health checkers for worker Valkey clients.

        Args:
            resource: The initialized Valkey clients

        Returns:
            Health checker for Valkey clients
        """
        return ValkeyHealthChecker(
            clients={
                CID_REDIS_LIVE: resource.valkey_live,
                CID_REDIS_STAT: resource.valkey_stat,
            }
        )



from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import Any, final

from aiohttp.streams import AsyncStreamIterator
from yarl import URL

from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.appproxy.common.types import SerializableCircuit as Circuit
from ai.backend.appproxy.worker.types import RootContext
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class HttpRequest:
    method: str
    path: str | URL
    headers: dict[str, Any]
    body: AsyncStreamIterator[bytes] | None


class BaseBackend:
    root_context: RootContext
    circuit: Circuit
    last_used: float

    def __init__(self, root_context: RootContext, circuit: Circuit) -> None:
        self.root_context = root_context
        self.circuit = circuit
        self.last_used = time.time()

    async def close(self) -> None:
        pass

    async def update_routes(self, routes: list[RouteInfo]) -> None:
        self.routes = routes

    @final
    async def mark_last_used_time(self, route: RouteInfo) -> None:
        await self.root_context.last_used_time_marker_redis_queue.put((
            [
                f"session.{route.session_id}.last_access",
                f"circuit.{self.circuit.id}.last_access",
            ],
            time.time(),
        ))

    @final
    async def increase_request_counter(self) -> None:
        await self.root_context.request_counter_redis_queue.put(
            f"circuit.{self.circuit.id}.requests"
        )



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path

from ai.backend.common.dependencies import DependencyComposer, DependencyStack

from .bootstrap import BootstrapComposer, BootstrapInput
from .infrastructure import InfrastructureComposer, InfrastructureResources


@dataclass
class DependencyInput:
    """Input required for complete dependency setup."""

    config_path: Path | None = None


@dataclass
class DependencyResources:
    """All dependency resources for app proxy worker."""

    infrastructure: InfrastructureResources


class WorkerDependencyComposer(DependencyComposer[DependencyInput, DependencyResources]):
    """Main composer for all app proxy worker dependencies."""

    @property
    def stage_name(self) -> str:
        return "worker"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: DependencyInput,
    ) -> AsyncIterator[DependencyResources]:
        """Compose all app proxy worker dependencies."""
        # Stage 1: Bootstrap (config)
        bootstrap = await stack.enter_composer(
            BootstrapComposer(),
            BootstrapInput(config_path=setup_input.config_path),
        )

        # Stage 2: Infrastructure (redis)
        infrastructure = await stack.enter_composer(
            InfrastructureComposer(),
            bootstrap.config,
        )

        yield DependencyResources(
            infrastructure=infrastructure,
        )



__all__ = [
    "TraefikPortFrontend",
    "TraefikSubdomainFrontend",
    "TraefikTCPFrontend",
    "HTTPPortFrontend",
    "HTTPSubdomainFrontend",
    "H2PortFrontend",
    "H2SubdomainFrontend",
    "TCPFrontend",
]

from .h2.port import PortFrontend as H2PortFrontend
from .h2.subdomain import SubdomainFrontend as H2SubdomainFrontend
from .http.port import PortFrontend as HTTPPortFrontend
from .http.subdomain import SubdomainFrontend as HTTPSubdomainFrontend
from .tcp import TCPFrontend
from .traefik import TraefikPortFrontend, TraefikSubdomainFrontend, TraefikTCPFrontend



import importlib.resources
import logging
import ssl

import aiohttp_jinja2
import jinja2
from aiohttp import web
from aiohttp.typedefs import Handler

from ai.backend.appproxy.common.exceptions import GenericBadRequest, ServerMisconfiguredError
from ai.backend.logging import BraceStyleAdapter

from ....config import WildcardDomainConfig
from ....types import Circuit, SubdomainFrontendInfo
from .base import BaseHTTPFrontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class SubdomainFrontend(BaseHTTPFrontend[str]):
    site: web.TCPSite | None
    wildcard_config: WildcardDomainConfig

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.site = None
        if not self.root_context.local_config.proxy_worker.wildcard_domain:
            raise ServerMisconfiguredError("worker:proxy-worker.wildcard-domain")
        self.wildcard_config = self.root_context.local_config.proxy_worker.wildcard_domain

    async def start(self) -> None:
        ssl_ctx = None
        proxy_worker_config = self.root_context.local_config.proxy_worker
        if proxy_worker_config.tls_listen:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(proxy_worker_config.tls_cert),
                str(proxy_worker_config.tls_privkey),
            )
        app = web.Application()
        app.on_response_prepare.append(self.append_cors_headers)
        app.middlewares.extend([
            self.ensure_slot_middleware,
            self.metric_collector_middleware,
            self.exception_middleware,
        ])
        app.router.add_route("*", "/{path:.*$}", self.proxy)

        with importlib.resources.as_file(
            importlib.resources.files("ai.backend.appproxy.common")
        ) as f:
            template_path = f / "templates"
            aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

        runner = web.AppRunner(app, keepalive_timeout=30.0)
        await runner.setup()
        service_addr = self.wildcard_config.bind_addr
        site = web.TCPSite(
            runner,
            str(service_addr.host),
            service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        await site.start()
        self.site = site
        log.info("accepting proxy requests at {}", service_addr)

    async def stop(self) -> None:
        if self.site:
            await self.site.stop()

    def parse_slot(self, request: web.Request) -> str:
        # exclude port number when evaluating the subdomain
        subdomain = request.host.partition(":")[0].replace(self.wildcard_config.domain, "").lower()
        if subdomain not in self.circuits:
            raise GenericBadRequest(f"E20009: Subdomain {subdomain} not registered")
        return subdomain

    @web.middleware
    async def ensure_slot_middleware(
        self, request: web.Request, handler: Handler
    ) -> web.StreamResponse:
        async def _exception_safe_handler(request: web.Request) -> web.StreamResponse:
            slot = self.parse_slot(request)
            circuit = self.circuits[slot]
            self.ensure_credential(request, circuit)
            backend = self.backends[slot]
            request["circuit"] = circuit
            request["backend"] = backend
            return await handler(request)

        return await self.exception_safe_handler_wrapper(request, _exception_safe_handler)

    def get_circuit_key(self, circuit: Circuit) -> str:
        assert isinstance(circuit.frontend, SubdomainFrontendInfo)
        return circuit.frontend.subdomain.lower()



import asyncio
import logging
import random
import socket
from typing import Final

import aiotools

from ai.backend.appproxy.common.exceptions import WorkerNotAvailable
from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.logging import BraceStyleAdapter

from .base import BaseBackend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

MAX_BUFFER_SIZE: Final[int] = 1 * 1024 * 1024


class TCPBackend(BaseBackend):
    routes: list[RouteInfo]

    def __init__(self, routes: list[RouteInfo], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.routes = routes

    @property
    def selected_route(self) -> RouteInfo:
        if len(self.routes) == 0:
            raise WorkerNotAvailable
        elif len(self.routes) == 1:
            selected_route = self.routes[0]
            if selected_route.traffic_ratio == 0:
                raise WorkerNotAvailable
        else:
            ratios: list[float] = [r.traffic_ratio for r in self.routes]
            selected_route = random.choices(self.routes, weights=ratios, k=1)[0]
        return selected_route

    async def bind(
        self, down_reader: asyncio.StreamReader, down_writer: asyncio.StreamWriter
    ) -> None:
        metrics = self.root_context.metrics
        total_bytes = 0
        stop_event = asyncio.Event()

        async def _pipe(
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            tag="(unknown)",
        ) -> None:
            nonlocal total_bytes

            try:
                while True:
                    data = await reader.read(n=MAX_BUFFER_SIZE)
                    if not data:
                        break
                    total_bytes += len(data)
                    metrics.proxy.observe_upstream_tcp_traffic_chunk(len(data))
                    writer.write(data)
                    await writer.drain()
                    log.debug("TCPBackend._pipe(t: {}): sent {} bytes", tag, len(data))
            except ConnectionResetError:
                log.debug("Conn reset")
                pass
            except Exception:
                log.exception("")
                raise
            finally:
                log.debug("setting stop event")
                stop_event.set()

        async def _last_access_marker_task(interval: float) -> None:
            await self.mark_last_used_time(route)

        route = self.selected_route
        log.debug(
            "Proxying TCP Request to {}:{}",
            route.current_kernel_host,
            route.kernel_port,
        )

        marker_task = aiotools.create_timer(_last_access_marker_task, 1.5)
        await self.increase_request_counter()

        try:
            sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # unlike .frontend.tcp this has a chance of being a blocking call since kernel host can be a domain
            await asyncio.get_running_loop().run_in_executor(
                None, sock.connect, (route.current_kernel_host, route.kernel_port)
            )

            up_reader, up_writer = await asyncio.open_connection(sock=sock)
            log.debug(
                "Connected to {}:{}",
                route.current_kernel_host,
                route.kernel_port,
            )
            async with asyncio.TaskGroup() as group:
                group.create_task(_pipe(up_reader, down_writer, tag="up->down"))
                group.create_task(_pipe(down_reader, up_writer, tag="down->up"))
        finally:
            log.debug("tasks ended")
            metrics.proxy.observe_upstream_tcp_traffic_chunk(total_bytes)
            marker_task.cancel()
            await marker_task
            down_writer.close()
            await down_writer.wait_closed()
        log.debug("TCP connection closed")



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.dependencies import DependencyComposer, DependencyStack

from ...config import ServerConfig
from .redis import RedisProvider, WorkerValkeyClients


@dataclass
class InfrastructureResources:
    """All infrastructure resources for app proxy worker."""

    valkey: WorkerValkeyClients


class InfrastructureComposer(DependencyComposer[ServerConfig, InfrastructureResources]):
    """Composer for infrastructure layer dependencies."""

    @property
    def stage_name(self) -> str:
        return "infrastructure"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: ServerConfig,
    ) -> AsyncIterator[InfrastructureResources]:
        """Compose all infrastructure dependencies."""
        # Setup infrastructure in dependency order
        valkey = await stack.enter_dependency(RedisProvider(), setup_input)

        yield InfrastructureResources(
            valkey=valkey,
        )



import importlib.resources
import logging
import ssl

import aiohttp_jinja2
import jinja2
from aiohttp import web
from aiohttp.typedefs import Handler

from ai.backend.appproxy.common.exceptions import GenericBadRequest, ServerMisconfiguredError
from ai.backend.logging import BraceStyleAdapter

from ....types import Circuit, PortFrontendInfo
from .base import BaseHTTPFrontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class PortFrontend(BaseHTTPFrontend[int]):
    sites: list[web.TCPSite]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.sites = []

    async def start(self) -> None:
        ssl_ctx = None
        proxy_worker_config = self.root_context.local_config.proxy_worker
        if proxy_worker_config.tls_listen:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(proxy_worker_config.tls_cert),
                str(proxy_worker_config.tls_privkey),
            )
        port_proxy_config = proxy_worker_config.port_proxy
        if not port_proxy_config:
            raise ServerMisconfiguredError("worker:proxy-worker.port-proxy")
        port_start, port_end = port_proxy_config.bind_port_range
        for port in range(port_start, port_end + 1):
            app = web.Application()
            app["port"] = port
            app.on_response_prepare.append(self.append_cors_headers)
            # ensure_slot middleware should be placed before exception_middleware and metric_collector_middleware
            # so that metric_collector_middleware can properly assess circuit variable
            # Keep in mind that implementation of ensure_slot_middleware should manually handle exception and respond with
            # appropriate HTTP response, which usually is automatically covered by exception_middleware
            # That can be achieved by wraping whole execution block with AbstractHTTPFrontend.exception_safe_handler_wrapper()
            app.middlewares.extend([
                self.ensure_slot_middleware,
                self.metric_collector_middleware,
                self.exception_middleware,
            ])
            app.router.add_route("*", "/{path:.*$}", self.proxy)

            with importlib.resources.as_file(
                importlib.resources.files("ai.backend.appproxy.common")
            ) as f:
                template_path = f / "templates"
                aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

            runner = web.AppRunner(app, keepalive_timeout=30.0)
            await runner.setup()
            service_host = port_proxy_config.bind_host
            site = web.TCPSite(
                runner,
                service_host,
                port,
                backlog=1024,
                reuse_port=True,
                ssl_context=ssl_ctx,
            )
            await site.start()
            self.sites.append(site)
        log.info(
            "accepting proxy requests from {}:{}~{}",
            port_proxy_config.bind_host,
            port_start,
            port_end,
        )

    async def stop(self) -> None:
        for site in self.sites:
            await site.stop()

    @web.middleware
    async def ensure_slot_middleware(
        self, request: web.Request, handler: Handler
    ) -> web.StreamResponse:
        port: int = request.app["port"]
        circuit = self.circuits.get(port, None)
        if circuit is None:
            raise GenericBadRequest(f"Unregistered slot {port}")
        self.ensure_credential(request, circuit)
        backend = self.backends[port]
        request["circuit"] = circuit
        request["backend"] = backend
        return await handler(request)

    def get_circuit_key(self, circuit: Circuit) -> int:
        assert isinstance(circuit.frontend, PortFrontendInfo)
        return circuit.frontend.port



from dataclasses import dataclass

import aiohttp
from yarl import URL

from .base import BaseBackend


@dataclass
class BackendConfig:
    hostname: str
    port: int
    domain: str | None = None


class H2Backend(BaseBackend):
    api_endpoint: URL

    def __init__(self, api_endpoint: URL, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.api_endpoint = api_endpoint

    async def update_config(self, backends: list[BackendConfig]) -> None:
        new_config = "\n".join([
            f"backend={b.hostname},{b.port};{b.domain or ''};proto=h2" for b in backends
        ])
        async with aiohttp.ClientSession(base_url=self.api_endpoint) as sess:
            async with sess.post("/api/v1beta1/backendconfig", data=new_config) as resp:
                resp.raise_for_status()



from __future__ import annotations

from .composer import BootstrapComposer, BootstrapInput, BootstrapResources

__all__ = [
    "BootstrapComposer",
    "BootstrapInput",
    "BootstrapResources",
]



from __future__ import annotations

import asyncio
import sys

import click

from ai.backend.common.dependencies.stacks.visualizing import VisualizingDependencyStack

from ..dependencies import DependencyInput, WorkerDependencyComposer


@click.group()
def cli() -> None:
    """Command set for dependency verification and validation."""
    pass


@cli.command(name="verify")
@click.option("--no-timestamps", is_flag=True, help="Hide timestamps in output")
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False),
    default="WARNING",
    help="Set the logging level (default: WARNING)",
)
def verify(
    no_timestamps: bool,
    log_level: str,
) -> None:
    """Verify all app proxy worker dependencies can be initialized successfully."""

    async def _verify_dependencies() -> bool:
        """Verify all app proxy worker dependencies can be initialized."""
        print("\n" + "=" * 60)
        print("App Proxy Worker Dependency Verification")
        print("=" * 60)
        print("Config: default locations")
        print(f"Log Level: {log_level.upper()}")
        print("=" * 60 + "\n")

        stack = VisualizingDependencyStack(
            output=sys.stdout,
            show_timestamps=not no_timestamps,
        )

        try:
            async with stack:
                dependency_input = DependencyInput(
                    config_path=None,
                )
                composer = WorkerDependencyComposer()

                await stack.enter_composer(composer, dependency_input)

        except Exception as e:
            print(f"\n❌ Dependency initialization failed: {e}\n")

        print()
        stack.print_summary()

        return not stack.has_failures()

    success = asyncio.run(_verify_dependencies())

    if success:
        print("All dependencies verified successfully!\n")
        sys.exit(0)
    else:
        print("Some dependencies failed. Check the output above for details.\n")
        sys.exit(1)



from ai.backend.appproxy.common.types import RouteInfo

from ...types import Circuit, RootContext
from .base import BaseBackend


class TraefikBackend(BaseBackend):
    worker_circuit: Circuit

    def __init__(self, context: RootContext, circuit: Circuit, routes: list[RouteInfo]) -> None:
        super().__init__(context, circuit)

        self.routes = routes
        self.worker_circuit = circuit



from __future__ import annotations

from .composer import InfrastructureComposer, InfrastructureResources
from .redis import RedisProvider, WorkerValkeyClients

__all__ = [
    "InfrastructureComposer",
    "InfrastructureResources",
    "RedisProvider",
    "WorkerValkeyClients",
]



from __future__ import annotations

import ipaddress
import os
import socket
import sys
import textwrap
from pathlib import Path
from pprint import pformat
from typing import Annotated, Self

import click
from pydantic import AnyUrl, Field, FilePath, ValidationError, model_validator

from ai.backend.appproxy.common.config import (
    BaseSchema,
    DebugConfig,
    GroupIDValidator,
    HostPortPair,
    PermitHashConfig,
    ProfilingConfig,
    RedisConfig,
    SecretConfig,
    UserIDValidator,
)
from ai.backend.appproxy.common.exceptions import ConfigValidationError
from ai.backend.appproxy.common.types import (
    AppMode,
    EventLoopType,
    FrontendMode,
    FrontendServerMode,
    ProxyProtocol,
)
from ai.backend.common import config
from ai.backend.common.types import ServiceDiscoveryType
from ai.backend.logging import LogLevel
from ai.backend.logging.config import LoggingConfig


def _get_default_uid() -> int:
    """Get UID from server.py file in the same directory as config.py."""
    server_file = Path(__file__).parent / "server.py"
    if not server_file.exists():
        raise ConfigValidationError(f"server.py not found at {server_file}")
    return server_file.stat().st_uid


def _get_default_gid() -> int:
    """Get GID from server.py file in the same directory as config.py."""
    server_file = Path(__file__).parent / "server.py"
    if not server_file.exists():
        raise ConfigValidationError(f"server.py not found at {server_file}")
    return server_file.stat().st_gid


class AppFilterConfig(BaseSchema):
    key: str
    value: str


class WildcardDomainConfig(BaseSchema):
    bind_addr: HostPortPair = Field(examples=[HostPortPair(host="127.0.0.1", port=10205)])
    domain: str = Field(
        description="Base domain for wildcard proxy.",
        examples=[".example.proxy.backend.ai"],
    )
    advertised_port: int | None = Field(
        default=None,
        description="Must be set if appproxy worker is between NAT.",
        examples=[10205],
    )


class PortProxyConfig(BaseSchema):
    bind_host: str = Field(examples=["127.0.0.1"])
    advertised_host: str | None = Field(default=None, examples=["127.0.0.1"])
    bind_port_range: tuple[int, int] = Field(examples=[(10205, 10300)])
    advertised_port_range: tuple[int, int] | None = Field(
        default=None,
        description="Must be set if appproxy worker is between NAT.",
        examples=[(10205, 10300)],
    )


class H2Config(BaseSchema):
    nghttpx_path: FilePath = Field(
        description="Path to nghttpx binary.",
        examples=[Path("/usr/local/bin/nghttpx")],
    )
    api_port_pool: tuple[int, int] = Field(
        default=(50000, 60000),
        description="Port pool for nghttpx API server.",
    )


class TraefikPortProxyConfig(BaseSchema):
    advertised_host: str = Field(examples=["127.0.0.1"])
    port_range: tuple[int, int] = Field(examples=[(10205, 10300)])


class TraefikWildcardDomainConfig(BaseSchema):
    domain: str = Field(
        description="Base domain for wildcard proxy.",
        examples=[".example.proxy.backend.ai"],
    )
    advertised_port: int
    tls_advertised: bool = Field(default=False)


class TraefikConfig(BaseSchema):
    api_port: int = Field(
        description="Port number of the `traefik` entrypoint binds.", examples=[8080]
    )
    frontend_mode: FrontendMode = Field(
        description="Type of frontend mode the worker will operate.",
        examples=[FrontendMode.WILDCARD_DOMAIN],
    )
    wildcard_domain: TraefikWildcardDomainConfig | None = Field(
        default=None,
        description="Must be filled if frontend_mode is 'wildcard'.",
    )
    port_proxy: TraefikPortProxyConfig | None = Field(
        default=None,
        description="Must be filled if frontend_mode is 'port'.",
    )
    last_used_time_marker_directory: Path = Field(
        examples=["/home/ubuntu/appproxy/worker-interactive"]
    )

    @model_validator(mode="after")
    def validate_mode_config(self) -> Self:
        match self.frontend_mode:
            case FrontendMode.PORT:
                if self.port_proxy is None:
                    raise ValueError("port_proxy must be set when frontend_mode = 'port'")
            case FrontendMode.WILDCARD_DOMAIN:
                if self.wildcard_domain is None:
                    raise ValueError("wildcard_domain must be set when frontend_mode = 'wildcard'")
        return self


class OTELConfig(BaseSchema):
    enabled: bool = Field(
        default=False,
        description=(
            "Whether to enable OpenTelemetry for tracing or logging. "
            "When enabled, traces or log will be collected and sent to the configured OTLP endpoint."
        ),
    )
    log_level: str = Field(
        default="INFO",
        description="Log level for OpenTelemetry logging.",
        examples=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    endpoint: str = Field(
        default="http://localhost:4317",
        description=(
            "OTLP (OpenTelemetry Protocol) endpoint for exporting telemetry data. "
            "Should include the host and port of the OTLP receiver."
        ),
        examples=["http://localhost:4317", "http://otel-collector:4317"],
    )


class ServiceDiscoveryConfig(BaseSchema):
    type: ServiceDiscoveryType = Field(
        default=ServiceDiscoveryType.REDIS,
        description="Type of service discovery to use.",
    )


class ProxyWorkerConfig(BaseSchema):
    ipc_base_path: Path = Field(
        default=Path("/tmp/backend.ai/ipc"),
        description="Directory to store temporary UNIX sockets.",
    )
    event_loop: EventLoopType = Field(
        default=EventLoopType.UVLOOP,
        description="Type of event loop to use.",
    )
    pid_file: Path = Field(
        default=Path(os.devnull),
        description="Place to store process PID.",
        examples=["/run/backend.ai/appproxy/coordinator.pid"],
    )

    id: str = Field(
        default=f"i-{socket.gethostname()}",
        description="Node id.",
    )
    user: Annotated[
        int,
        UserIDValidator,
        Field(default_factory=_get_default_uid, ge=0, description="Process owner."),
    ]
    group: Annotated[
        int,
        GroupIDValidator,
        Field(default_factory=_get_default_gid, ge=0, description="Process group."),
    ]

    api_bind_addr: HostPortPair = Field(
        default=HostPortPair(host="0.0.0.0", port=10201),
        description="Hostname and port number worker API server will listen to.",
    )
    api_advertised_addr: HostPortPair | None = Field(
        default=None,
        description="Hostname and port number which API users can access.",
        examples=[HostPortPair(host="127.0.0.1", port=10201)],
    )

    coordinator_endpoint: AnyUrl = Field(
        description="HTTP(S) URI to coordinator API endpoint.",
        examples=["http://127.0.0.1:10200"],
    )
    verify_coordinator_ssl_certificate: bool = Field(
        default=True,
        description="Validate coordinator's SSL certificate",
    )

    authority: str = Field(
        description=(
            "Unique, human-readable appproxy worker identifier. "
            "Must be set equal across every worker nodes representing same worker port via High-Availability setup."
        ),
        examples=["worker-1"],
    )

    use_experimental_redis_event_dispatcher: bool = Field(
        default=False,
        description="Use experimental Redis event dispatcher implementation.",
    )

    tls_listen: bool = Field(default=False, description="Opt in to HTTPS setup.")
    tls_cert: FilePath | None = Field(
        default=None,
        description="Path to TLS certificate PEM.",
        examples=["/etc/backend.ai/tls/cert.pem"],
    )
    tls_privkey: FilePath | None = Field(
        default=None,
        description="Path to TLS private key PEM.",
        examples=["/etc/backend.ai/tls/privkey.pem"],
    )
    tls_advertised: bool = Field(
        default=False,
        description="Must be active if proxy coordinator is served behind external TLS terminator.",
    )

    aiomonitor_termui_port: int = Field(
        default=48600,
        description="Port number for aiomonitor termui server.",
        gt=0,
        lt=65536,
    )
    aiomonitor_webui_port: int = Field(
        default=49600,
        description="Port number for aiomonitor webui server.",
        gt=0,
        lt=65536,
    )

    heartbeat_period: float = Field(
        description="Heartbeat period in seconds.",
        default=10.0,
        gt=0,
    )

    frontend_mode: FrontendServerMode = Field(
        description="Type of frontend mode the worker will operate.",
    )
    protocol: ProxyProtocol = Field(
        description="Type of protocol worker will serve. `HTTP2` not allowed here.",
    )
    wildcard_domain: WildcardDomainConfig | None = Field(
        default=None, description="Must be filled if frontend_mode is 'wildcard'."
    )
    port_proxy: PortProxyConfig | None = Field(
        default=None, description="Must be filled if frontend_mode is 'port'."
    )
    traefik: TraefikConfig | None = Field(
        default=None, description="Must be filled if frontend_mode is 'traefik'."
    )
    http2: H2Config | None = Field(
        default=None, description="Must be filled if protocol is 'http2'."
    )

    accepted_traffics: list[AppMode] = Field(
        description="Limit only selected kind of traffics to walk through this worker.",
    )
    app_filters: list[AppFilterConfig] = Field(
        default_factory=list,
        description="Define app filters.",
        examples=[
            [AppFilterConfig(key="session.id", value="CED03BE1-3ABE-4FAB-A23B-5CC9ABC60A04")]
        ],
    )
    filtered_apps_only: bool = Field(
        default=False,
        description="If active, only apps matching defined filters will be allowed to be proxied by this worker.",
    )

    metric_access_allowed_hosts: str = Field(
        default="127.0.0.1/32",
        description="Limits access to `/metrics` HTTP resources from described hosts only.",
    )

    inference_metric_collection_interval: float = Field(
        default=5.0,
        description="The interval of inference metric collection (secs)",
    )

    client_pool_cleanup_interval: float = Field(
        default=60.0,
        description="The interval to sweep unused aiohttp.ClientSession instances to make backend requests to kernel apps (secs)",
    )

    announce_addr: HostPortPair | None = Field(
        default=None,
        description=(
            "The announced address for service discovery. "
            "If not set, it will use api_bind_addr for announcement."
        ),
    )

    @model_validator(mode="after")
    def populate_announce_addr(self) -> Self:
        if self.announce_addr is None:
            self.announce_addr = self.api_bind_addr
        return self

    @model_validator(mode="after")
    def validate_metric_access_allowed_hosts(self) -> Self:
        try:
            ipaddress.IPv4Network(self.metric_access_allowed_hosts)
        except ValueError:
            raise ValueError(
                "metric_access_allowed_hosts should be either a valid IPv4 Address or Network"
            ) from None
        return self

    @model_validator(mode="after")
    def validate_mode_config(self) -> Self:
        match self.frontend_mode:
            case FrontendServerMode.PORT:
                if self.port_proxy is None:
                    raise ValueError("port_proxy must be set when frontend_mode = 'port'")
            case FrontendServerMode.WILDCARD_DOMAIN:
                if self.wildcard_domain is None:
                    raise ValueError("wildcard_domain must be set when frontend_mode = 'wildcard'")
            case FrontendServerMode.TRAEFIK:
                if self.traefik is None:
                    raise ValueError("traefik must be set when frontend_mode = 'traefik'")
        return self

    @model_validator(mode="after")
    def validate_protocol_config(self) -> Self:
        match self.protocol:
            case ProxyProtocol.GRPC:
                raise ValueError(
                    "Directly setting protocol = 'grpc' is not supported; use 'h2' instead"
                )
            case ProxyProtocol.HTTP2:
                if not self.http2:
                    raise ValueError("http2 config must be set when protocol = 'h2'")
        return self


class ServerConfig(BaseSchema):
    redis: RedisConfig
    proxy_worker: ProxyWorkerConfig
    profiling: ProfilingConfig = Field(default_factory=lambda: ProfilingConfig())
    secrets: SecretConfig
    permit_hash: PermitHashConfig
    logging: LoggingConfig
    debug: DebugConfig
    otel: OTELConfig = Field(
        default_factory=lambda: OTELConfig(
            enabled=False,
            log_level="INFO",
            endpoint="http://localhost:4317",
        )
    )
    service_discovery: ServiceDiscoveryConfig = Field(
        default_factory=lambda: ServiceDiscoveryConfig(type=ServiceDiscoveryType.REDIS)
    )


def load(config_path: Path | None = None, log_level: LogLevel = LogLevel.NOTSET) -> ServerConfig:
    # Determine where to read configuration.
    raw_cfg, _ = config.read_from_file(config_path, "app-proxy-worker")

    config.override_key(raw_cfg, ("debug", "enabled"), log_level == LogLevel.DEBUG)
    if log_level != LogLevel.NOTSET:
        config.override_key(raw_cfg, ("logging", "level"), log_level)
        config.override_key(raw_cfg, ("logging", "pkg-ns", "ai.backend"), log_level)

    # Validate and fill configurations
    # (allow_extra will make configs to be forward-copmatible)
    try:
        server_config = ServerConfig.model_validate(raw_cfg)
        if server_config.profiling.enable_pyroscope:
            if not server_config.profiling.pyroscope_config:
                raise ConfigValidationError("Pyroscope enabled but config is not populated")
            if server_config.profiling.pyroscope_config.application_name is None:
                server_config.profiling.pyroscope_config.application_name = f"proxy-worker-{server_config.proxy_worker.authority}-{server_config.proxy_worker.api_bind_addr.port}"
    except (ValidationError, ConfigValidationError) as e:
        print(
            "ConfigurationError: Could not read or validate the manager local config:",
            file=sys.stderr,
        )
        if isinstance(e, ValidationError):
            detail = str(e)
        else:
            detail = pformat(e)
        print(textwrap.indent(detail, prefix="  "), file=sys.stderr)
        raise click.Abort()
    else:
        if server_config.debug.enabled:
            print("== Proxy Worker configuration ==", file=sys.stderr)
            print(pformat(server_config.model_dump()), file=sys.stderr)
        return server_config
